import pickle
from collections import defaultdict
from math import ceil

from contracts import contract
from .constants import ETERNITY

__all__ = [
    "ExecutionStats",
    "write_stats",
]


class Statistics:
    def __init__(self, block):
        self.block = block
        self.num = 0
        self.mean_cpu = 0
        self.var_cpu = 0
        self.mean_wall = 0
        self.var_wall = 0
        self.perc_cpu = None
        self.perc_wall = None
        self.perc_times = None
        self.baseline_fraction = None
        self.last_timestamp = None
        self.mean_delta = 0
        self.delta_num = 0


class ExecutionStats(object):
    def __init__(self):
        self.samples = {}

        self.connection2samples = defaultdict(lambda: [])
        self.block2samples = defaultdict(lambda: [])

    @contract(connection="isinstance(BlockConnection)", timestamp="pg_timestamp_or_eternity")
    def add_connection_sample(self, connection, value, timestamp):
        if timestamp == ETERNITY:
            timestamp = 0.0
        else:
            timestamp = float(timestamp)

        from .model import BlockConnection

        assert isinstance(connection, BlockConnection)
        # warnings.warn("Size computation disabled due to speed")
        if self.connection2samples[connection]:
            s = self.connection2samples[connection][0]["size"]
        else:

            s = len(pickle.dumps(value))
        sample = dict(timestamp=float(timestamp), size=s)
        self.connection2samples[connection].append(sample)

    @contract(timestamp="pg_timestamp_or_eternity")
    def add_signal_sample(self, block, cpu, wall, timestamp):
        # assert cpu >= 0, cpu
        sample = dict(timestamp=timestamp, cpu=cpu, wall=wall)
        self.block2samples[block].append(sample)

    def get_all(self):
        def summarize(values):
            perc = [5, 25, 50, 75, 95]

            import numpy as np

            x = list(map(float, np.percentile(values, perc)))
            return dict(list(zip(perc, x)))

        def summarize_samples(samples):
            res = {}
            fields = list(samples[0].keys())
            for f in fields:
                values = [s[f] for s in samples]
                res[f] = summarize(values)
            return res

        res = {}
        res["blocks"] = {}
        from procgraph.core.model import Model

        for block, samples in list(self.block2samples.items()):
            if not isinstance(block, Model):
                res["blocks"][block.name] = dict(stats=summarize_samples(samples))

        res["signals"] = []
        for connection, samples in list(self.connection2samples.items()):
            stats = summarize_samples(samples)
            v = {
                "stats": stats,
                "block1": connection.block1.name,
                "block1_signal": connection.block1_signal,
            }

            if isinstance(connection.block2, Model):
                # if connected to a model, we are really connecting to the input
                block2_inputblock = connection.block2.model_input_ports[connection.block2_signal]
                block2_inputblock_name = "%s.%s" % (connection.block2.name, block2_inputblock.name)
                v["block2_signal"] = connection.block2_signal
                v["block2"] = block2_inputblock_name
            else:
                v["block2_signal"] = connection.block2_signal
                v["block2"] = connection.block2.name

            if isinstance(connection.block1, Model):
                # if connected to a model, we are really connecting to the input
                block1_inputblock = connection.block1.model_output_ports[connection.block1_signal]
                block1_inputblock_name = "%s.%s" % (connection.block1.name, block1_inputblock.name)
                v["block1_signal"] = connection.block1_signal
                v["block1"] = block1_inputblock_name
            else:
                v["block1_signal"] = connection.block1_signal
                v["block1"] = connection.block1.name
            res["signals"].append(v)
        return res

    #### Old interface

    def add(self, block, cpu, wall, timestamp):
        # weird wall/cpu behavior when suspending the process
        self.add_signal_sample(block, cpu, wall, timestamp)

        if wall < 0:
            wall = 1
        if cpu < 0:
            cpu = wall

        if cpu == 0:
            cpu = 0.0001
        if wall == 0:
            wall = 0.0001

        if not block in self.samples:
            self.samples[block] = Statistics(block)

        s = self.samples[block]

        # TODO: set parameter
        WINDOW = 100
        N = min(s.num, WINDOW)

        s.mean_cpu = (s.mean_cpu * N + cpu) / (N + 1)
        s.var_cpu = (s.var_cpu * N + (cpu - s.mean_cpu) ** 2) / (N + 1)
        s.mean_wall = (s.mean_wall * N + wall) / (N + 1)
        s.var_wall = (s.var_wall * N + (wall - s.mean_wall) ** 2) / (N + 1)
        s.num += 1

        if isinstance(timestamp, float):
            if s.last_timestamp is not None and s.last_timestamp != "eternity":
                if isinstance(s.last_timestamp, str):
                    raise ValueError(s.last_timestamp)
                delta = timestamp - s.last_timestamp
                # when the block is a model, it could be called multiple times
                if delta > 0:
                    N = s.delta_num
                    s.mean_delta = (s.mean_delta * N + delta) / (N + 1)
                    s.delta_num += 1
                    # print "%s %d mean %d" %
                    # (s.block, delta * 1000, 1000 * s.mean_delta)

        s.last_timestamp = timestamp

    def print_info(self):
        samples = list(self.samples.values())
        write_stats(samples)


def write_stats(samples):
    for s in samples:
        assert isinstance(s, Statistics)

    # get the block that executed fewest times and use it as a baseline
    baseline = min(samples, key=lambda s: s.num)

    # update percentage
    total_cpu = sum([s.mean_cpu * s.num for s in samples])
    total_wall = sum([s.mean_wall * s.num for s in samples])
    total_times = sum([s.num for s in samples])
    for s in samples:
        s.perc_cpu = s.mean_cpu * s.num / total_cpu
        s.perc_wall = s.mean_wall * s.num / total_wall
        s.perc_times = s.num * 1.0 / total_times
        s.baseline_fraction = s.num * 1.0 / baseline.num

    # sort by percentage
    alls = sorted(list(samples), key=lambda x: (-x.perc_wall))
    min_perc = 3
    print(
        (
            "--- Statistics (ignoring < %d%%) baseline: %d iterations of %s"
            % (min_perc, baseline.num, baseline.block)
        )
    )
    for s in alls:
        perc_cpu = ceil(s.perc_cpu * 100)
        perc_wall = ceil(s.perc_wall * 100)
        if s != baseline and (perc_cpu < min_perc) and (perc_wall < min_perc):
            continue
        perc_times = ceil(s.perc_times * 100)

        # jitter_cpu = ceil(100 * (sqrt(s.var_cpu) * 2 / s.mean_cpu))
        # jitter_wall = ceil(100 * (sqrt(s.var_wall) * 2 / s.mean_wall))

        if s.mean_cpu < 0.7 * s.mean_wall:
            comment = "IO "
        else:
            comment = "   "
        # print ''.join([
        # '- cpu: %dms (+-%d%%) %02d%% of total; ' %
        # (1000 * s.mean_cpu, jitter_cpu, perc_cpu),
        # 'wall: %dms (+-%d%%) %02d%% of total; ' %
        # (1000 * s.mean_wall, jitter_wall, perc_wall),
        # 'exec: %02d%% of total' % perc_times])

        if s.mean_delta > 0:
            fps = 1.0 / s.mean_delta
            stats = "%3.1f fps" % fps
        else:
            stats = " " * len("%3.1f fps" % 0)

        name = str(s.block)
        if len(name) > 35:
            name = name[:35]
        print(
            (
                "".join(
                    [
                        " cpu %4dms %2d%%| " % (1000 * s.mean_cpu, perc_cpu),
                        "wall %4dms %2d%%| " % (1000 * s.mean_wall, perc_wall),
                        "exec %2d%% %3d (%3.1fx) %s| %s "
                        % (perc_times, s.num, s.baseline_fraction, stats, comment),
                        name,
                    ]
                )
            )
        )



from contextlib import contextmanager
import os
import gzip


def is_gzip_filename(filename):
    return ".gz" in filename


@contextmanager
def safe_write(filename, mode="wb", compresslevel=5):
    """ 
        Makes atomic writes by writing to a temp filename. 
        Also if the filename ends in ".gz", writes to a compressed stream.
        Yields a file descriptor.
    """
    from procgraph.block_utils.file_io import make_sure_dir_exists

    make_sure_dir_exists(filename)

    tmp_filename = "%s.tmp" % filename
    try:
        if is_gzip_filename(filename):
            fopen = lambda f, mode: gzip.open(filename=f, mode=mode, compresslevel=compresslevel)
        else:
            fopen = open

        with fopen(tmp_filename, mode) as f:
            yield f

        os.rename(tmp_filename, filename)
    except:
        if os.path.exists(tmp_filename):
            os.unlink(tmp_filename)
        if os.path.exists(filename):
            os.unlink(filename)
        raise


@contextmanager
def safe_read(filename, mode="rb"):
    """ 
        If the filename ends in ".gz", reads from a compressed stream.
        Yields a file descriptor.
    """
    try:
        if is_gzip_filename(filename):
            fopen = gzip.open
        else:
            fopen = open

        with fopen(filename, mode) as f:
            yield f
    except:
        # TODO
        raise



from procgraph import Block
from procgraph.block_utils import make_sure_dir_exists
import os


PROCGRAPH_LOG_GROUP = "procgraph"


class BagWrite(Block):
    """ This block writes the incoming signals to a ROS bag file.

    By default, the signals are organized as follows: ::

         /            (root)
         /procgraph             (group with name procgraph)
         /procgraph/signal1     (table)
         /procgraph/signal2     (table)
         ...


    Note that the input should be ROS messages; no conversion is done.

    """

    Block.alias("bagwrite")
    Block.input_is_variable("Signals to be written", min=1)
    Block.config("file", "Bag file to write")

    def init(self):
        from ros import rosbag  # @UnresolvedImport

        self.info(f"Writing to bag {self.config.file!r}.")
        make_sure_dir_exists(self.config.file)

        if os.path.exists(self.config.file):
            os.unlink(self.config.file)
        self.tmp_file = self.config.file + ".active"
        self.bag = rosbag.Bag(self.tmp_file, "w", compression="bz2")
        self.signal2last_timestamp = {}

    def finish(self):
        self.bag.close()
        os.rename(self.tmp_file, self.config.file)

    def update(self):
        import rospy

        signals = self.get_input_signals_names()

        for signal in signals:
            msg = self.get_input(signal)
            timestamp = self.get_input_timestamp(signal)
            if (signal in self.signal2last_timestamp) and (timestamp == self.signal2last_timestamp[signal]):
                continue
            self.signal2last_timestamp[signal] = timestamp

            ros_timestamp = rospy.Time.from_sec(timestamp)
            self.bag.write("/%s" % signal, msg, ros_timestamp)



from procgraph import register_model_spec

# XXX: make models directory

register_model_spec(
    """
--- model covariance
''' Computes the covariance matrix of the input. '''
input x 'Unidimensional numpy array.'
output cov_x 'Square matrix representing sample covariance.'
config wait=10 'Number of sample to have reliable expectation.'

|input name=x| --> x --> |expectation| --> |wait n=$wait| --> Ex

   x, Ex --> |sync| --> |-| --> x_normalized 
   
   x_normalized, x_normalized --> |outer| -->  |expectation| --> covariance 
   
   covariance --> |output name=cov_x|
    
"""
)

register_model_spec(
    """
--- model normalize 
''' Removes the mean from a signal. '''
input  x       'Unidimensional numpy array.'
output x_n     'Signal without the mean.'
config wait=10 'Number of sample to have reliable expectation.'

|input name=x| --> x --> |expectation| --> |wait n=$wait| --> Ex

   x, Ex --> |sync| --> |-| --> x_normalized --> |output name=x_n|
    
"""
)



from contracts import contract
import datetime
import time

TIMESTAMP_FIELD = "timestamp"


@contract(timestamp="float", returns="str")
def iso_from_timestamp(timestamp):
    dt = datetime.datetime.fromtimestamp(timestamp)
    iso = str(dt)
    iso = iso.replace(" ", "T")
    return iso


@contract(iso="str", returns="float")
def timestamp_from_iso(iso):
    """ Returns a timestamp in double """
    import dateutil.parser

    d = dateutil.parser.parse(iso)  # datetime
    s = time.mktime(d.timetuple()) + 0.001 * 0.001 * d.microsecond
    return s



""" YAML conversions. """

procgraph_info = {"requires": ["yaml"]}


from procgraph import import_magic

yaml = import_magic(__name__, "yaml")


from .yaml2object import *
from .object2yaml import *

__all__ = ["yaml2object"]



from pyparsing import ParserElement
from procgraph.core.parsing_elements import parse_input_port, parse_output_port

# Enable memoization; much faster, but we can't use
# parse actions with side effects.
ParserElement.enablePackrat()

from pyparsing import (
    Regex,
    Word,
    delimitedList,
    alphas,
    Optional,
    OneOrMore,
    stringEnd,
    alphanums,
    ZeroOrMore,
    Group,
    Suppress,
    lineEnd,
    Combine,
    nums,
    Literal,
    CaselessLiteral,
    restOfLine,
    QuotedString,
    ParseException,
    Forward,
)

from .parsing_elements import (
    VariableReference,
    ParsedBlock,
    ParsedAssignment,
    ImportStatement,
    ParsedModel,
    ParsedSignal,
    ParsedSignalList,
    Connection,
    Where,
    output_from_tokens,
    input_from_tokens,
    config_from_tokens,
)
from .exceptions import PGSyntaxError


def eval_dictionary(s, loc, tokens):  # @UnusedVariable
    if not "content" in tokens:
        return {}
    d = {}
    for a in tokens:
        if "value" in a:
            d[a["key"]] = a["value"]
    return d


def eval_array(s, loc, tokens):  # @UnusedVariable
    elements = tokens.asList()
    res = []
    for i in range(len(elements)):
        t = elements[i]
        res.append(t)
    return res


# Shortcuts
S = Suppress
O = Optional

# Important: should be at the beginning
# make end of lines count
ParserElement.setDefaultWhitespaceChars(" \t")

# These are just values
# Definition of numbers
number = Word(nums)
point = Literal(".")
e = CaselessLiteral("E")
plusorminus = Literal("+") | Literal("-")
integer = Combine(O(plusorminus) + number)
# Note that '42' is not a valid float...
floatnumber = Combine(integer + point + O(number) + O(e + integer)) | Combine(integer + e + integer)

ellipsis_value = Literal("...").setParseAction(lambda _: Ellipsis)


def convert_int(tokens):
    assert len(tokens) == 1
    return int(tokens[0])


integer.setParseAction(convert_int)
floatnumber.setParseAction(lambda tokens: float(tokens[0]))

# comments
comment = S(Literal("#") + restOfLine)
good_name = Combine(Word(alphas) + O(Word(alphanums + "_")))

# All kinds of python strings

single_quoted = QuotedString('"', "\\", unquoteResults=True) | QuotedString("'", "\\", unquoteResults=True)
multi_quoted = QuotedString(
    quoteChar='"""', escChar="\\", multiline=True, unquoteResults=True
) | QuotedString(quoteChar="'''", escChar="\\", multiline=True, unquoteResults=True)
quoted = multi_quoted | single_quoted

reference = Combine(S("$") + good_name("variable"))

reference.setParseAction(VariableReference.from_tokens)

dictionary = Forward()
array = Forward()

value = (
    ellipsis_value
    | quoted
    | array
    | dictionary
    | reference
    | good_name
    | floatnumber  # order is important...
    | integer
)("val")

# dictionaries

dict_key = good_name | quoted
dictionary << (
    S("{") + O(delimitedList(Group(dict_key("key") + S(":") + value("value"))))("content") + S("}")
)


dictionary.setParseAction(eval_dictionary)

array << Group(S("[") + O(delimitedList(value)("elements")) + S("]"))

array.setParseAction(eval_array)


def parse_value(string, filename=None):
    """ This is useful for debugging """

    try:
        ret_value = value.parseString(string, parseAll=True)
        return ret_value["val"]

    except ParseException as e:
        where = Where(filename, string, line=e.lineno, column=e.col)
        raise PGSyntaxError("Error in parsing string: %s" % e, where=where)


def create_model_grammar():

    # We pass a "where" object to the constructors
    def wrap(constructor):
        def from_tokens(string, location, tokens):
            element = constructor(tokens)
            element.where = Where(ParsedModel.static_filename, string, location)
            return element

        return from_tokens

    arrow = S(Regex(r"-+>"))
    newline = S(lineEnd)

    # (don't put '.' at the beginning)
    qualified_name = Combine(good_name + "." + (integer | good_name))

    block_name = good_name
    block_type = good_name | Word("_+-/*") | quoted | reference

    signal = (
        O(S("[") + (integer | good_name)("local_input") + S("]"))
        + O(block_name("block_name") + S("."))
        + (integer | good_name)("name")
        + O(S("[") + (integer | good_name)("local_output") + S("]"))
    )
    signal.setParseAction(wrap(ParsedSignal.from_tokens))

    signals = delimitedList(signal)
    signals.setParseAction(wrap(ParsedSignalList.from_tokens))

    # Note that here the order matters (as qualified = good + something)
    key = qualified_name | good_name

    key_value_pair = Group(key("key") + S("=") + value("value"))

    parameter_list = OneOrMore(key_value_pair)
    parameter_list.setParseAction(lambda s, l, t: dict([(a[0], a[1]) for a in t]))  # @UnusedVariable

    block = (
        S("|")
        + O(block_name("name") + S(":"))
        + block_type("blocktype")
        + O(parameter_list("config"))
        + S("|")
    )
    block.setParseAction(wrap(ParsedBlock.from_tokens))

    input_port = S("(") + block_name("port_name") + S(")")
    output_port = S("(") + block_name("port_name") + S(")")
    output_port.setParseAction(wrap(parse_output_port))
    input_port.setParseAction(wrap(parse_input_port))

    between1 = arrow + O(signals + arrow)
    between2 = arrow + S(newline + arrow) + O(signals + arrow)
    between3 = newline + arrow + O(signals + arrow)
    between = between3 | between2 | between1

    # Different patterns

    # a -> |b| -> c
    arrow_arrow = signals + arrow + O(block + ZeroOrMore(between + block)) + arrow + signals
    # a  -> c
    arrow_arrow2 = signals + arrow + signals
    source = (input_port | block) + ZeroOrMore(between + block) + arrow + signals
    sink = signals + arrow + ZeroOrMore(block + between) + (block | output_port)
    # |a| -> |b| -> |c|
    source_sink = block + ZeroOrMore(between + block)
    # (in) -> |b| -> |c|
    source_sink2 = input_port + ZeroOrMore(between + block)
    # |a| ->  |b| -> (out)
    source_sink3 = ZeroOrMore(block + between) + output_port
    # (in) -> |a| -> (out)
    source_sink4 = input_port + between + ZeroOrMore(block + between) + output_port

    # all of those are called a "connection"
    connection = (
        arrow_arrow ^ sink ^ source ^ source_sink ^ source_sink2 ^ source_sink3 ^ source_sink4 ^ arrow_arrow2
    )

    connection.setParseAction(wrap(Connection.from_tokens))

    # allow breaking lines with backslash
    continuation = "\\" + lineEnd
    connection.ignore(continuation)

    assignment = key("key") + S("=") + value("value")
    assignment.setParseAction(wrap(ParsedAssignment.from_tokens))

    package_name = good_name + ZeroOrMore("." + good_name)
    import_statement = S("import") + package_name("package")
    import_statement.setParseAction(wrap(ImportStatement.from_tokens))

    config = S("config") + good_name("variable") + O(S("=") + value("default")) + O(quoted("docstring"))
    config.setParseAction(wrap(config_from_tokens))

    input_s = S("input") + good_name("name") + O(quoted("docstring"))
    input_s.setParseAction(wrap(input_from_tokens))

    output = S("output") + good_name("name") + O(quoted("docstring"))
    output.setParseAction(wrap(output_from_tokens))

    docs = S(ZeroOrMore(multi_quoted + OneOrMore(newline)))

    action = (
        comment
        | config
        | input_s
        | output
        | (docs + connection)
        | (docs + assignment)
        | (docs + import_statement)
    )

    model_content = (
        ZeroOrMore(newline) + action + ZeroOrMore(OneOrMore(newline) + action) + ZeroOrMore(newline)
    )

    named_model = (
        S(Combine("---" + O(Word("-"))))
        + S("model")
        + good_name("model_name")
        + OneOrMore(newline)
        + O(quoted("docstring"))
        + model_content("content")
    )

    named_model.setParseAction(wrap(ParsedModel.from_named_model))

    anonymous_model = model_content.copy()
    anonymous_model.setParseAction(wrap(ParsedModel.from_anonymous_model))

    comments = ZeroOrMore((comment + newline) | newline)
    pg_file = comments + (OneOrMore(named_model) | anonymous_model) + stringEnd  # important

    return pg_file


pg_file = create_model_grammar()


def parse_model(string, filename=None):
    """ Returns a list of ParsedModel """
    # make this check a special case, otherwise it's hard to debug
    if not string.strip():
        msg = "Passed empty string."
        raise PGSyntaxError(msg, Where(filename, string, 0))

    # this is not threadsafe (but we don't have threads, so it's all good)
    ParsedModel.static_filename = filename

    try:
        parsed = pg_file.parseString(string)
        return list(parsed)
    except ParseException as e:
        where = Where(filename, string, line=e.lineno, column=e.col)
        msg = "Error in parsing string: %s" % e
        raise PGSyntaxError(msg, where=where)


# Register handler for ellipsis, so that ParsedModel is pickable.
# Ellipsis is they only object we use that is not pickable by default.
import copyreg


def code_unpickler(data):  # @UnusedVariable
    return Ellipsis


def code_pickler(code):  # @UnusedVariable
    return code_unpickler, (None,)


copyreg.pickle(type(Ellipsis), code_pickler, code_unpickler)



from procgraph import register_simple_block
from procgraph_images.copied_from_reprep import skim_top_and_bottom


def skim(a, percent=5):

    return skim_top_and_bottom(a, percent)


register_simple_block(skim, doc=skim_top_and_bottom.__doc__)

# doc='Skims the top and bottom percentile from the data.')



from optparse import OptionParser
from setuptools import find_packages
from pprint import pprint

from .pgdoc import get_all_info
from ..core.registrar import default_library
from ..core.import_magic import REQUIRES_PARSED


def main():
    parser = OptionParser()

    parser.add_option("--output", default="procgraph_packages.py")
    parser.add_option("--src", default="src", help="Source directory")

    (options, args) = parser.parse_args()

    if args:
        raise Exception("No arguments necessary.")

    given_modules = find_packages(options.src)
    given_modules.remove("procgraph.block_utils")
    given_modules.remove("procgraph.core")
    given_modules.remove("procgraph.scripts")
    given_modules.remove("procgraph.testing")

    for module in given_modules:
        __import__(module, fromlist=["ceremony"])

    all_modules = get_all_info(default_library)

    # check they were not empty
    all_info = {"packages": {}}
    for module in given_modules:
        if not module in all_modules:
            print(
                (
                    "Warning: I found no blocks defined *directly* in %r. "
                    "No index will be generated." % module
                )
            )
            continue

        info = all_modules[module]
        extract = {}
        # extract['name'] = info.name
        extract["desc"] = info.desc
        extract["blocks"] = list(info.blocks.keys())
        extract["requires"] = info.procgraph_info.get(REQUIRES_PARSED, {})
        all_info["packages"][info.name] = extract

    print(("Writing to %r." % options.output))
    with open(options.output, "w") as f:
        f.write("# Autogenerated file -- do not modify.\n\n")
        f.write('# This file is generated by "pgindex".  \n')
        f.write(
            "# It lists the default ProcGraph packages and their dependencies \n"
            "# so that we can warn the user if some dependency is not found at \n"
            "# install time. ProcGraph's policy is to be very liberal; for example, \n"
            "# the `procgraph_cv' package is installed even though OpenCV is not found.\n"
            "# and an error will be thrown only if the user actually uses those blocks.\n"
            "\n"
        )
        f.write("index = ")
        pprint(all_info, f)


if __name__ == "__main__":
    main()



from . import cv  # @UnresolvedImport
from .conversions import cv_to_numpy
from procgraph import Generator, Block
import time


class CVCapture(Generator):

    Block.alias("cv_capture")
    Block.config("cam", default=0)
    Block.config("width", default=160)
    Block.config("height", default=120)
    Block.config("fps", default=10)

    Block.output("rgb")

    def init(self):
        cam = self.config.cam
        width = self.config.width
        height = self.config.height
        fps = self.config.fps

        self.info("Capturing cam=%d %dx%d @%.1f fps" % (cam, width, height, fps))
        self.capture = cv.CaptureFromCAM(cam)
        cv.SetCaptureProperty(self.capture, cv.CV_CAP_PROP_FRAME_WIDTH, width)
        cv.SetCaptureProperty(self.capture, cv.CV_CAP_PROP_FRAME_HEIGHT, height)
        cv.SetCaptureProperty(self.capture, cv.CV_CAP_PROP_FPS, fps)

    def update(self):
        t = time.time()
        img = cv.QueryFrame(self.capture)
        rgb = cv_to_numpy(img)
        self.set_output("rgb", value=rgb, timestamp=t)

    def next_data_status(self):
        return (True, None)



from procgraph import simple_block, COMPULSORY

from .pil_conversions import Image_from_array


__all__ = ["imwrite"]


@simple_block
def imwrite(rgb, file=COMPULSORY):  # @ReservedAssignment
    """ 
        Writes an image to a file.
    """
    im = Image_from_array(rgb)
    im.save(file)



import numpy as np

from procgraph import Block


class Join(Block):
    """ 
    This block joins multiple signals into one.
    """

    Block.alias("join")

    Block.input_is_variable("Signals to be joined together.")
    Block.output("joined", "Joined signals.")

    def init(self):
        sizes = {}
        names = self.get_input_signals_names()
        for signal in names:
            sizes[signal] = None

        self.state.sizes = sizes

    def update(self):
        sizes = self.state.sizes
        result = []
        for name in self.get_input_signals_names():
            value = self.get_input(name)
            # workaround for scalar values
            value = np.reshape(value, np.size(value))
            size = len(value)
            if value is None:
                return
            if sizes[name] is None:
                sizes[name] = size
            else:
                if size != sizes[name]:
                    raise Exception("Signal %s changed size from %s to %s." % (name, sizes[name], size))

            result.extend(value)

        self.output[0] = np.array(result)



from contracts import contract, new_contract


class DrawToken:
    def __init__(self, position, text):
        self.position = position
        self.text = text

    def __repr__(self):
        return "DT([%1.f,%1.f], %r)" % (self.position[0], self.position[1], self.text)


new_contract("drawtoken", DrawToken)


def tokenize(s):
    s = s.replace("\n", " ")
    tokens = s.split(" ")
    tokens = [t for t in tokens if t]
    #     print('tokenize(%s) -> %s' % (s,tokens))
    return tokens


@contract(text=str, line_width=">0", line_scale=">1", returns="tuple(list(drawtoken), tuple(float,float))")
def format_paragraph(text, font, line_width, line_scale=1.2):
    """ Returns a list of drawtokens and width/height """
    # print('format_paragraph %r' % text)
    tokens = tokenize(text)

    def size_token(token):
        w, h = font.getsize(token)
        # print('  token %r size h%s w%s font %r' % (token, h, w, font))
        return w * 1.0, h * 1.0

    def width_from_token(token):
        w, _ = size_token(token)
        return w

    space_width, _ = size_token("n")
    _, line_height0 = size_token("lgH")
    extra_between_lines = line_height0 * (line_scale - 1)
    extra_between_lines = 0.0
    line_height = line_height0 + extra_between_lines

    # print('line width (max): %s' % line_width)
    print(f"line height: {line_height}")
    print(f"extra: {extra_between_lines}")
    # print('space_width: %s' % space_width)
    lines = split_in_lines(tokens, width_from_token, line_width, space_width)
    # print('Obtained %d lines: %s' % (len(lines), lines))
    out = []
    max_w = 0.0
    for i, l in enumerate(lines):
        if len(l) == 0:
            print(("warning: empty line in %r" % lines))
        cur_y = i * line_height
        dts, w = format_line(line=l, width_from_token=width_from_token, ybase=cur_y, space_width=space_width)
        if w > line_width:
            msg = "Line %r longer than allowed (%s > %s)" % (l, w, line_width)
            from procgraph.core.visualization import error

            error(msg)

        max_w = max(max_w, w)
        out.extend(dts)

    height = len(lines) * line_height + (len(lines) - 1) * extra_between_lines
    return out, (height, max_w)


@contract(line="list(str)", returns="tuple(list(drawtoken), >=0)")
def format_line(line, width_from_token, ybase, space_width):
    out = []
    cur_x = 0.0
    for token in line:
        # print('cur: %s token: %r w: %r' % (cur_x, token, wt))
        o = DrawToken(position=(ybase, cur_x), text=token)
        out.append(o)
        wt = width_from_token(token)
        cur_x += wt + space_width

    if line:
        cur_x -= space_width
    return out, cur_x


@contract(returns="list(list(str))")
def split_in_lines(tokens, width_from_token, line_width, space_width):
    tokens = list(tokens)
    lines = []
    cur_line = []
    cur_line_width = 0.0
    for token in tokens:
        token_width = width_from_token(token)
        # print(' %r -> %r' % (token, token_width))
        too_long = cur_line_width + token_width + space_width > line_width
        # begin new line if we go over, and we put at least one
        begin_new = too_long and len(cur_line) > 0
        if begin_new:
            lines.append(cur_line)
            cur_line_width = 0.0
            cur_line = []
        cur_line.append(token)
        cur_line_width += token_width + space_width
    if cur_line:
        lines.append(cur_line)
    # print('split_in_lines(tokens=%r) -> %r' % (tokens, lines))
    return lines



import subprocess
import yaml
from contracts import contract
from procgraph.utils.inawhile import InAWhile
import warnings
from contracts.utils import indent
from procgraph import logger


def rosbag_info(bag):
    warnings.warn("remove, use the other version")
    stdout = subprocess.Popen(["rosbag", "info", "--yaml", bag], stdout=subprocess.PIPE).communicate()[0]
    try:
        info_dict = yaml.load(stdout)
    except:
        logger.error("Could not parse yaml:\n%s" % indent(stdout, "| "))
        raise
    return info_dict


# Example output:
# path: /home/andrea/01-youbot-ros-logs-good/unicornA_base1_2013-04-02-20-37-43.bag
# version: 2.0
# duration: 3499.176103
# start: 1364960263.862855
# end: 1364963763.038959
# size: 997946578
# messages: 2656954
# indexed: True
# compression: bz2
# uncompressed: 1840288997
# compressed: 962203048
# types:
#     - type: array_msgs/FloatArray
#       md5: 788898178a3da2c3718461eecda8f714
# topics:
#     - topic: /arm_1/arm_controller/position_command
#       type: brics_actuator/JointPositions

from procgraph import logger as pg_logger


def read_bag_stats_progress(source, logger, interval=5):
    warnings.warn("remove, use the other version")

    tracker = InAWhile(interval)

    for topic, msg, t, extra in source:
        if tracker.its_time():
            progress = extra["t_percentage"]
            cur_obs = extra["counter"]
            num_obs = "?"  # extra['messages']  # XXX: this is all messages
            status = "%4.1f%%  (obs: %4d/%s);  %5.1f fps" % (progress, cur_obs, num_obs, tracker.fps())
            logger.debug(status)
        yield topic, msg, t, extra


@contract(topics="list(str)")
def read_bag_stats(bagfile, topics, logger=None):
    """ 
        This yields a dict with: 
        
            topic
            msg
            t
            
            info:
                counter
                messages

                start
                end
                t_percentage
                t_from_start
            
            
    
    """
    warnings.warn("remove, use the other version")

    import rosbag  # @UnresolvedImport
    from rospy import rostime  # @UnresolvedImport

    if logger is None:
        logger = pg_logger

    logger.debug("Reading info for bagfile...")
    bag_info = rosbag_info(bagfile)
    logger.debug("Opening bagfile...")
    bag = rosbag.Bag(bagfile)

    extra = bag_info
    extra["messages"] = bag_info["messages"]
    extra["start"] = bag_info["start"]
    extra["end"] = bag_info["end"]
    extra["duration"] = bag_info["duration"]
    t0 = rostime.Time.from_sec(bag_info["start"])
    i = 0
    for topic, msg, t in bag.read_messages(topics=topics):
        if i == 0:
            logger.debug("first message arrived.")
        extra["counter"] = i
        extra["t_from_start"] = (t - t0).to_sec()
        extra["t_percentage"] = 100.0 * (extra["t_from_start"] / extra["duration"])
        i += 1
        yield topic, msg, t, extra



from procgraph import Block

__all__ = ["AllReady"]


class AllReady(Block):
    """ 
        This block outputs the inputs, unchanged. 
    """

    Block.alias("all_ready")

    Block.input_is_variable("Input signals.", min=1)
    Block.output_is_variable("Output signals, equal to input.")

    def update(self):
        if not self.all_input_signals_ready():
            return

        for i in range(self.num_input_signals()):
            if self.input_update_available(i):
                t, value = self.get_input_ts_and_value(i)
                self.set_output(i, value, t)



"""
    Utilities for writing compact file paths.
"""
__version__ = "1.0"
import os

# TODO: cache the results?


def friendly_path(path, use_environment=True):
    """
        Gets a friendly representation of the given path,
        using relative paths or environment variables
        (if use_environment = True).
    """
    original = path
    options = []

    options.append(os.path.relpath(path, os.getcwd()))

    rules = []
    rules.append(("~", os.path.expanduser("~")))

    if use_environment:
        envs = dict(os.environ)
        # remove unwanted
        for e in list(envs.keys()):
            if "PWD" in e:
                del envs[e]

        for k, v in list(envs.items()):
            if v:
                if v and v[-1] == "/":
                    v = v[:-1]
                rules.append(("${%s}" % k, v))

    # apply longest first
    rules.sort(key=lambda x: (-len(x[1])))
    path = replace_variables(path, rules)

    options.append(path)

    weight_doubledot = 5

    def score(s):
        # penalize '..' a lot
        s = s.replace("..", "*" * weight_doubledot)
        return len(s)

    options.sort(key=score)

    if False:
        print(("Options for %s" % original))
        for o in options:
            print(("- %4d %s" % (score(o), o)))

    result = options[0]

    # print('Converted %s  => %s' % (original, result))

    return result


def replace_variables(path, rules):
    for k, v in rules:
        if path.startswith(v):
            # print("  applied %s => %s" % (v, k))
            path = path.replace(v, k)
    return path



from .utils import PGTestCase

good_examples = [" y -> |$block| -> x ", ' y -> |"block"| -> x ', " y -> |block| -> x "]


bad_examples = [" y -> |1| -> x ", " y -> |[1,2]| -> x "]


class SyntaxTest(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from . import cv, np
from .conversions import numpy_to_cv
from procgraph import Block
import warnings


class Display(Block):
    Block.alias("cv_display")

    Block.config("name", default=None)
    Block.config("position", default=None)

    Block.input("rgb")

    nimages = 0

    def init(self):
        name = self.config.name
        if name is None:
            name = "display%d" % Display.nimages
        self.name = name

        Display.nimages += 1

        cv.NamedWindow(self.name, 1)

        if self.config.position is not None:
            x, y = self.config.position
        else:
            cols = 4
            w, h = 320, 320
            u = Display.nimages % cols
            v = int(np.floor(Display.nimages / cols))
            x = u * w
            y = v * h

        cv.MoveWindow(self.name, x, y)

    def update(self):
        rgb = self.input.rgb
        img = numpy_to_cv(rgb)
        cv.ShowImage(self.name, img)

    def finish(self):
        warnings.warn("to fix")
        cv.DestroyAllWindows()



from procgraph import Block


class FPSPrint(Block):
    """ Prints the fps count for the input signals. """

    Block.alias("fps_print")

    Block.input_is_variable("Any signal.", min=1)

    def init(self):
        self.state.last_timestamp = None

    def update(self):
        current = max(self.get_input_signals_timestamps())
        last = self.state.last_timestamp
        if last is not None:

            difference = current - last
            fps = 1.0 / difference

            self.info("FPS %s %.1f" % (self.canonicalize_input(0), fps))

        self.state.last_timestamp = current



""" 
    Blocks using the OpenCV library. 
"""


procgraph_info = {
    # List of python packages
    "requires": [("cv2", ("cv2",)), "numpy"]
}


from procgraph import import_magic, import_successful

cv = import_magic(__name__, "cv2", "cv")
# if not import_successful(cv):
#     cv = import_magic(__name__, 'cv2')
np = import_magic(__name__, "numpy")

from .opencv_utils import *
from .cv_capture import *
from .cv_display import *
from .detection import *

if not import_successful(cv):
    if False:
        from procgraph import logger

        logger.warn("Could not import CV")


from procgraph import pg_add_this_package_models

pg_add_this_package_models(__file__, __package__)



from procgraph import Block

__all__ = ["WaitSec"]


class WaitSec(Block):
    """ 
        This block only transmits for t > t0.
        If retime is true, then the timestamp starts at 0.
    """

    Block.alias("wait_sec")

    Block.config("t0", "Delta to start.")
    Block.config("retime", "If true, retime from t=0", default=False)

    Block.input("x")
    Block.output("y")

    def init(self):
        #         self.state.t0 = None
        pass

    def update(self):

        t = self.get_input_timestamp(0)
        #         if self.state.t0 is None:
        #             self.state.t0 = t
        #
        delta = t - self.config.t0
        #
        #         self.info('config.t0 %.3f t %.3f delta %.3f' % (      self.config.t0,
        #                                                               delta,  t))

        if delta > 0:
            if self.config.retime:
                t_ = delta
            else:
                t_ = t
            #             self.info('t_: %s' % t_)
            self.set_output(0, value=self.get_input(0), timestamp=t_)


#         else:
#             self.info('skip')



from procgraph.core.model_loader import model_from_string

from .utils import PGTestCase


class NamingTest(PGTestCase):
    def test_correct_naming(self):
        # in all these cases, the identity block should have
        # the output signal named "x"
        examples = [
            " |input name=x| -> |identity| ",
            " |input name=y| -> x -> |identity| ",
            """
          |i0:input name=y| -> x

          x -> |identity|
        """,
            """
          |i0:input name=y|

          i0.y[x] -> |identity|
        """,
        ]

        for example in examples:
            self.try_one(example)

    def try_one(self, example):
        model = model_from_string(example)
        block = model.name2block["identity"]
        names = block.get_output_signals_names()
        print(("Obtained %s from '''%s'''" % (names, example)))
        self.assertEqual(len(names), 1)
        self.assertEqual(names[0], "x")



import numpy as np
from numpy.ma.testutils import assert_equal

from procgraph import BadConfig, BadInput, Block
from . import pylab, pylab2rgb
from .fanciness import fancy_styles

try:
    from time import thread_time as measure_thread_time
except ImportError:
    from time import clock as measure_thread_time

__all__ = [
    "Plot",
]


class Plot(Block):
    """
        Plots the inputs using matplotlib.

        This block accepts an arbitrary number of signals.
        Each signals is treated independently and plot separately.

        Each signal can either be:

        1.  A tuple of length 2. It is interpreted as a tuple ``(x,y)``,
            and we plot ``x`` versus ``y`` (see also :ref:`block:make_tuple`).

        2.  A list of numbers, or a 1-dimensional numpy array of length N.
            In this case, it is interpreted as the y values,
            and we set  ``x = 1:N``.

    """

    Block.alias("plot")

    Block.config("width", "Image dimension", default=320)
    Block.config("height", "Image dimension", default=240)
    Block.config("xlabel", "X label for the plot.", default=None)
    Block.config("ylabel", "Y label for the plot.", default=None)
    Block.config("legend", "List of strings to use as legend handles.", default=None)
    Block.config("title", 'If None, use the signal name. Set to ``""`` to disable.', default=None)
    Block.config("format", 'Line format ("-",".","x-",etc.)', default="-")
    Block.config(
        "symmetric",
        "An alternative to y_min, y_max." " Makes sure the plot is symmetric for y. ",
        default=False,
    )
    Block.config("x_min", "If set, force the X axis to have this minimum.", default=None)
    Block.config("x_max", "If set, force the X axis to have this maximum.", default=None)
    Block.config("y_min", "If set, force the Y axis to have this minimum.", default=None)
    Block.config("y_max", "If set, force the Y axis to have this maximum.", default=None)
    Block.config(
        "keep",
        "If True, tries to reuse the figure, without closing." " (buggy on some backends)",
        default=False,
    )
    Block.config("transparent", "If true, outputs a RGBA image instead of RGB.", default=False)

    Block.config(
        "tight", 'Uses "tight" option for creating png (Matplotlib>=1.1).', default=False,
    )
    Block.config(
        "fancy_styles", "A list of fancy styles to apply (%s)." % list(fancy_styles.keys()), default=[]
    )

    Block.input_is_variable("Data to plot.")

    Block.output("rgb", "Resulting image.")

    def init(self):
        self.line = None

        # figure gets initialized in update() on the first execution
        self.figure = None

        self.warned = False

    def init_figure(self):
        width = self.config.width
        height = self.config.height

        # TODO: remove from here
        pylab.rc("xtick", labelsize=8)
        pylab.rc("ytick", labelsize=8)

        """ Creates figure object and axes """
        self.figure = pylab.figure(frameon=False, figsize=(width / 100.0, height / 100.0))
        # TODO: is there a better way?
        # left, bottom, right, top
        # borders = [0.15, 0.15, 0.03, 0.05]
        # w = 1 - borders[0] - borders[2]
        # h = 1 - borders[1] - borders[3]
        # self.axes = pylab.axes([borders[0], borders[1], w, h])
        self.axes = pylab.axes()
        self.figure.add_axes(self.axes)

        pylab.draw_if_interactive = lambda: None

        pylab.figure(self.figure.number)
        if self.config.title is not None:
            if self.config.title != "":
                self.axes.set_title(self.config.title, fontsize=10)
        else:
            # We don't have a title ---
            t = ", ".join(self.get_input_signals_names())
            self.axes.set_title(t, fontsize=10)

        if self.config.xlabel:
            self.axes.set_xlabel(self.config.xlabel)
        if self.config.ylabel:
            self.axes.set_ylabel(self.config.ylabel)

        self.legend_handle = None

        self.lines = {}
        self.lengths = {}

    def apply_styles(self, pylab):
        # TODO: check type
        for style in self.config.fancy_styles:
            if not style in fancy_styles:
                error = "Cannot find style %r in %s" % (style, list(fancy_styles.keys()))
                raise BadConfig(error, self, "style")
            function = fancy_styles[style]
            function(pylab)

    def plot_one(self, id, x, y, format):  # @ReservedAssignment
        assert isinstance(x, np.ndarray)
        assert isinstance(y, np.ndarray)
        assert len(x.shape) <= 1
        assert len(y.shape) <= 1
        assert len(x) == len(y)

        if id in self.lengths:
            if self.lengths[id] != len(x):
                redraw = True
                self.axes.lines.remove(self.lines[id])
            else:
                redraw = False
        else:
            redraw = True

        if redraw:
            res = self.axes.plot(x, y, format)
            line = res[0]

            self.lines[id] = line
        else:
            self.lines[id].set_ydata(y)
            self.lines[id].set_xdata(x)

        self.lengths[id] = len(x)

        if self.limits is None:
            self.limits = np.array([min(x), max(x), min(y), max(y)])
        else:
            self.limits[0] = min(self.limits[0], min(x))
            self.limits[1] = max(self.limits[1], max(x))
            self.limits[2] = min(self.limits[2], min(y))
            self.limits[3] = max(self.limits[3], max(y))

        # self.limits = map(float, self.limits)

    def update(self):
        self.limits = None

        start = measure_thread_time()

        if self.figure is None:
            self.init_figure()

        pylab.figure(self.figure.number)

        self.apply_styles(pylab)

        for i in range(self.num_input_signals()):
            value = self.input[i]
            if value is None:
                raise BadInput("Input is None (did you forget a |sync|?)", self, i)
            elif isinstance(value, tuple):
                if len(value) != 2:
                    raise BadInput("Expected tuple of length 2 instead of %d." % len(value), self, i)

                xo = value[0]
                yo = value[1]

                if xo is None or yo is None:
                    raise BadInput("Invalid members of tuple", self, i)

                x = np.array(xo)
                y = np.array(yo)

                # X must be one-dimensional
                if len(x.shape) > 1:
                    raise BadInput("Bad x vector w/shape %s." % str(x.shape), self, i)

                # y should be of dimensions ...?
                if len(y.shape) > 2:
                    raise BadInput("Bad shape for y vector %s." % str(y.shape), self, i)

                if len(x) != y.shape[0]:
                    raise BadInput(
                        "Incompatible dimensions x: %s, y: %s" % (str(x.shape), str(y.shape)), self, i
                    )
                # TODO: check x unidimensional (T)
                # TODO: check y compatible dimensions (T x N)

            else:
                y = np.array(value)
                if len(y.shape) > 2:
                    raise BadInput("Bad shape for y vector %s." % str(y.shape), self, i)

                if len(y.shape) == 1:
                    x = np.array(list(range(len(y))))
                else:
                    assert len(y.shape) == 2
                    x = np.array(list(range(y.shape[1])))

            if len(x) <= 1:
                continue

            if len(y.shape) == 2:
                y = y.transpose()

            if len(y.shape) == 1:
                pid = self.canonicalize_input(i)
                self.plot_one(pid, x, y, self.config.format)
            else:
                assert len(y.shape) == 2
                num_lines = y.shape[0]
                for k in range(num_lines):
                    pid = "%s-%d" % (self.canonicalize_input(i), k)
                    yk = y[k, :]
                    self.plot_one(pid, x, yk, self.config.format)
            # TODO: check that if one has time vector, also others have it

        if self.limits is not None:

            if self.config.x_min is not None:
                self.limits[0] = self.config.x_min
            if self.config.x_max is not None:
                self.limits[1] = self.config.x_max
            if self.config.y_min is not None:
                self.limits[2] = self.config.y_min
            if self.config.y_max is not None:
                self.limits[3] = self.config.y_max

            if self.config.symmetric:
                if self.config.y_min is not None or self.config.y_max is not None:
                    raise BadConfig(
                        "Cannot specify symmetric together with" "y_min or y_max.", self, "symmetric"
                    )

                M = max(abs(self.limits[2:4]))
                if "dottedzero" in self.config.fancy_styles:
                    a = pylab.axis()
                    pylab.plot([a[0], a[1]], [0, 0], "k--")

                self.limits[2] = -M
                self.limits[3] = M

            # leave some space above and below
            self.limits[2] = self.limits[2] * 1.1
            self.limits[3] = self.limits[3] * 1.1

            self.axes.axis(self.limits)

        if self.legend_handle is None:
            legend = self.config.legend
            if legend:
                self.legend_handle = self.axes.legend(
                    *legend,
                    loc="upper right",
                    handlelength=1.5,
                    markerscale=2,
                    labelspacing=0.03,
                    borderpad=0,
                    handletextpad=0.03,
                    borderaxespad=1,
                )

        # http://matplotlib.sourceforge.net/users/tight_layout_guide.html
        try:
            pylab.tight_layout()
        except Exception as e:
            msg = "Could not call tight_layout(); available only on " "Matplotlib >=1.1 (%s)" % e
            if not self.warned:
                self.warning(msg)
                self.warned = True

        plotting = measure_thread_time() - start

        start = measure_thread_time()
        # There is a bug that makes the image smaller than desired
        # if tight is True
        pixel_data = pylab2rgb(transparent=self.config.transparent, tight=self.config.tight)

        # So we check and compensate
        shape = pixel_data.shape[0:2]
        shape_expected = (self.config.height, self.config.width)
        from procgraph_images import image_pad  # need here, otherwise circular

        if shape != shape_expected:
            msg = "pylab2rgb() returned size %s instead of %s." % (shape, shape_expected)
            msg += " I will pad the image with white."
            self.warning(msg)
            pixel_data = image_pad(pixel_data, shape_expected, bgcolor=[1, 1, 1])
            assert_equal(pixel_data.shape[0:2], shape_expected)

        reading = measure_thread_time() - start

        if False:
            # 30 49 30 before
            print(("plotting: %dms    reading: %dms" % (plotting * 1000, reading * 1000)))

        self.output.rgb = pixel_data

        if not self.config.keep:
            pylab.close(self.figure.number)
            self.figure = None



from . import np


from procgraph import simple_block, COMPULSORY
from procgraph.block_utils import check_2d_array

# TODO: add for pycontracts
# add_contract('color', 'list[3](>=0,<=1)')


@simple_block
def skim_top(a, top_percent=COMPULSORY):
    """ Cuts off the top percentile of the array.
    
        :param top_percent: How much to cut off (decimal).
         :type top_percent: float,>=0,<90
    """
    assert top_percent >= 0 and top_percent < 90
    threshold = np.percentile(a.flat, 100 - top_percent)
    return np.minimum(a, threshold)


@simple_block
def skim_top_and_bottom(a, percent=COMPULSORY):
    """ Cuts off the top and bottom percentile of the array. 
    
        :param a: Any numpy array. 
         :type a: array[x]
            
        :param percent: How much to cut off (decimal).
         :type percent: float,>=0,<90

        :return: Skimmed version of ``a``.
         :rtype: array[x]
        
    """
    assert percent >= 0 and percent < 90
    threshold_max = np.percentile(a.flat, 100 - percent)
    threshold_min = np.percentile(a.flat, percent)
    return np.maximum(threshold_min, np.minimum(a, threshold_max))


@simple_block
def posneg(value, max_value=None, skim=0, nan_color=[0.5, 0.5, 0.5], zero_color=[1.0, 1.0, 1.0]):
    """ 
        Converts a 2D float value to a RGB representation, where
        red is positive, blue is negative, white is zero.
    
        :param value: The field to represent.
         :type value: array[HxW]
        
        :param max_value:  Maximum of absolute value (if None, detect).
         :type max_value:  float,>0
        
        :param skim:       Fraction to skim (in percent).
         :type skim:       float,>0,<100
         
        :param nan_color:  Color to give for regions of NaN and Inf.
         :type nan_color:  color
        
        :return: posneg: A RGB image.
         :rtype: array[HxWx3](uint8)

    """

    check_2d_array(value, "input to posneg")

    # TODO: put this in reprep
    value = value.copy()
    if value.ndim > 2:
        value = value.squeeze()

    if value.dtype == np.dtype("uint8"):
        value = value.astype("float32")

    if len(value.shape) != 2:
        raise Exception("I expected a H x W image, got shape %s." % str(value.shape))

    isfinite = np.isfinite(value)
    isnan = np.logical_not(isfinite)
    # set nan to 0
    value[isnan] = 0

    if max_value is None:
        abs_value = abs(value)
        if skim != 0:
            abs_value = skim_top(abs_value, skim)

        max_value = np.max(abs_value)

        if max_value == 0:
            result = np.zeros((value.shape[0], value.shape[1], 3), dtype="uint8")
            for i in range(3):
                result[:, :, i] = zero_color[i] * 255
            return result

    assert np.isfinite(max_value)

    positive = np.minimum(np.maximum(value, 0), max_value) / max_value
    negative = np.maximum(np.minimum(value, 0), -max_value) / -max_value
    positive_part = (positive * 255).astype("uint8")
    negative_part = (negative * 255).astype("uint8")

    result = np.zeros((value.shape[0], value.shape[1], 3), dtype="uint8")

    anysign = np.maximum(positive_part, negative_part)
    R = 255 - negative_part[:, :]
    G = 255 - anysign
    B = 255 - positive_part[:, :]

    # remember the nans
    R[isnan] = nan_color[0] * 255
    G[isnan] = nan_color[1] * 255
    B[isnan] = nan_color[2] * 255

    result[:, :, 0] = R
    result[:, :, 1] = G
    result[:, :, 2] = B

    return result


@simple_block
def scale(
    value, min_value=None, max_value=None, min_color=[1, 1, 1], max_color=[0, 0, 0], nan_color=[1, 0, 0]
):
    """ 
        Provides a RGB representation of the values by interpolating the range 
        [min(value),max(value)] into the colorspace [min_color, max_color].
        
        :param value: The field to represent.
         :type value: array[HxW],H>0,W>0
            
        :param max_value: If specified, everything *above* is clipped.
         :type max_value: float
        
        :param min_value: If specified, everything *below* is clipped.
         :type min_value: float
    
        :param min_color:  Color to give to the minimum values.
         :type min_color:  color
        :param max_color:  Color to give to the maximum values.
         :type max_color:  color
        :param nan_color:  Color to give for regions of NaN and Inf.
         :type nan_color:  color
        
        :return: scale: A RGB image.
        :rtype: array[HxWx3](uint8)

    """
    # Raises :py:class:`.ValueError` if min_value == max_value

    check_2d_array(value, "input to scale()")

    # assert_finite(value)
    value = value.copy()
    if value.ndim > 2:
        value = value.squeeze()

    if value.dtype == np.dtype("uint8") or value.dtype == np.dtype("int"):
        value = value.astype("float32")
    # require_shape((gt(0), gt(0)), value)

    min_color = np.array(min_color)
    max_color = np.array(max_color)
    nan_color = np.array(nan_color)
    # require_shape((3,), min_color)
    # require_shape((3,), max_color)

    isnan = np.logical_not(np.isfinite(value))

    if max_value is None:
        value[isnan] = -np.inf
        max_value = np.max(value)

    if min_value is None:
        value[isnan] = np.inf
        min_value = np.min(value)

    if max_value == min_value or np.isnan(min_value) or np.isnan(max_value):
        # XXX: maybe allow set this case as an exception?
        #        raise ValueError('I end up with max_value = %s = %s = min_value.' % \
        #                         (max_value, min_value))
        result = np.zeros((value.shape[0], value.shape[1], 3), dtype="uint8")
        result[:, :, :] = 255
        return result

    value01 = (value - min_value) / (max_value - min_value)

    # Cut at the thresholds
    value01 = np.maximum(value01, 0)
    value01 = np.minimum(value01, 1)

    result = np.zeros((value.shape[0], value.shape[1], 3), dtype="uint8")

    for u in [0, 1, 2]:
        col = 255 * ((1 - value01) * min_color[u] + (value01) * max_color[u])
        col[isnan] = nan_color[u] * 255
        result[:, :, u] = col

    return result



from .utils import PGTestCase

good_examples = [
    '|verify  x="a" y="a"|',
    """
    var=value
    |verify  x=$var y="value"|""",
    """
    var=value
    |verify  x="${var}" y="value"|""",
    """
    var=value
    |verify  x="${var}" y=$var|""",
    """
    var1=value
    var2=$var1
    |verify  x="${var2}" y="value"|""",
    """
    var1=value
    var2=$var1
    |verify  x="1${var2}2" y="1value2"|""",
    """
    var1=value
    |verify  x={a: $var1} y={a: value}|""",
    """
    var1=value
    |verify  x=[$var1] y=[value]|""",
]


bad_examples = [
    '|verify  x="b" y="a"|',
    # Recursive cases
    """
    one = $one
""",
    """
    one = "${one}"
""",
    """
    one = $two
    two = $one
""",
    """
    one = "${two}"
    two = $one
""",
    """
    two = $one
    one = "${two}"
""",
    """
    one = "${two}"
    two = "${one}"
""",
]


class SubstitutionTest(PGTestCase):
    def testGoodExamples(self):
        for example in good_examples:
            self.check_semantic_ok(example)

    def testBadExamples(self):
        for example in bad_examples:
            self.check_semantic_error(example)



# str -> function f(pyplot)
fancy_styles = {}


def style(f):
    fancy_styles[f.__name__] = f
    return f


@style
def dickinsonA(pylab):
    set_spines_look_A(pylab)


def remove_spine(pylab, which):
    ax = pylab.gca()
    ax.spines[which].set_color("none")


@style
def notopaxis(pylab):
    remove_spine(pylab, "top")


@style
def nobottomaxis(pylab):
    remove_spine(pylab, "bottom")


@style
def noxticks(pylab):
    pylab.xticks([], [])


@style
def noyticks(pylab):
    pylab.yticks([], [])


@style
def turn_off_right(pylab):
    ax = pylab.gca()
    for loc, spine in list(ax.spines.items()):
        if loc in ["right"]:
            spine.set_color("none")  # don't draw spine
    ax.yaxis.set_ticks_position("left")


@style
def turn_off_left_right(pylab):
    ax = pylab.gca()
    for loc, spine in list(ax.spines.items()):
        if loc in ["right", "left"]:
            spine.set_color("none")  # don't draw spine
    ax.yaxis.set_ticks_position("none")


@style
def turn_off_top(pylab):
    ax = pylab.gca()
    for loc, spine in list(ax.spines.items()):
        if loc in ["top"]:
            spine.set_color("none")  # don't draw spine
    ax.xaxis.set_ticks_position("bottom")


@style
def spines_outward(pylab, offset=10):
    ax = pylab.gca()
    for loc, spine in list(ax.spines.items()):
        if loc in ["left", "bottom", "top", "right"]:
            spine.set_position(("outward", offset))


@style
def turn_off_all_axes(pylab):
    turn_off_bottom_and_top(pylab)
    turn_off_left_right(pylab)


@style
def turn_off_bottom_and_top(pylab):
    ax = pylab.gca()
    for loc, spine in list(ax.spines.items()):
        if loc in ["bottom", "top"]:
            spine.set_color("none")  # don't draw spine

    pylab.xticks([], [])


def set_spines_outward(pylab, outward_offset=10):
    ax = pylab.gca()
    for loc, spine in list(ax.spines.items()):
        if loc in ["left", "bottom"]:
            spine.set_position(("outward", outward_offset))
        elif loc in ["right", "top"]:
            spine.set_color("none")  # don't draw spine
        else:
            raise ValueError("unknown spine location: %s" % loc)

    # turn off ticks where there is no spine
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")


def set_spines_look_A(pylab, outward_offset=10, linewidth=2, markersize=3, markeredgewidth=1):
    """
        Taken from
        http://matplotlib.sourceforge.net/examples/pylab_examples
        /spine_placement_demo.html
    """
    set_spines_outward(pylab, outward_offset=outward_offset)
    ax = pylab.gca()

    for l in ax.get_xticklines() + ax.get_yticklines():
        l.set_markersize(markersize)
        l.set_markeredgewidth(markeredgewidth)

    try:
        ax.get_frame().set_linewidth(linewidth)
    except BaseException as e:
        print("set_linewidth() not working in matplotlib 1.3.1")
        print(e)


@style
def dottedzero(pylab):
    pass


#     a = pylab.axis()
#     pylab.plot([a[0], a[1]], [0, 0], '--')



from procgraph import Block


class FPSDataLimit(Block):
    """ This block limits the output update to a certain framerate. """

    Block.alias("fps_data_limit")

    Block.config("fps", "Maximum framerate.")

    Block.input_is_variable("Signals to decimate.", min=1)
    Block.output_is_variable("Decimated signals.")

    def init(self):
        self.state.last_timestamp = None

    def update(self):
        should_update = False

        last = self.state.last_timestamp
        current = max(self.get_input_signals_timestamps())

        if last is None:
            should_update = True
            self.state.last_timestamp = current
        else:
            fps = self.config.fps
            delta = 1.0 / fps
            difference = current - last
            if difference > delta:
                should_update = True
                self.state.last_timestamp = current

        if not should_update:
            return
        # Just copy the input to the output
        for i in range(self.num_input_signals()):
            self.set_output(i, self.get_input(i), self.get_input_timestamp(i))



from procgraph import simple_block


@simple_block
def crop(rgb, left=0, right=0, top=0, bottom=0):
    """ Crops an image by the given values"""

    # print('Cropping %s' % (str(rgb.shape)))
    height = rgb.shape[0]
    width = rgb.shape[1]

    rgb2 = rgb[top : (height - bottom), left : (width - right), :]

    if rgb2.shape[0] == 0 or rgb2.shape[1] == 0:
        msg = "Obtained empty image (t %s, r %s, b %s, l %s) on (%s,%s)" % (
            top,
            right,
            bottom,
            left,
            height,
            width,
        )
        raise ValueError(msg)

    #     print('cropping %s -> %s' % (str(rgb.shape), str(rgb2.shape)))

    #    if left > 0:
    #        rgb = rgb[:, left:, :]
    #
    #    if right > 0:
    #        rgb = rgb[:, :-right, :]
    #
    #    if top > 0:
    #        rgb = rgb[top:, :, :]
    #
    #    if bottom > 0:
    #        rgb = rgb[:-bottom, :, :]
    return rgb2



from procgraph import simple_block
from . import np

__all__ = ["reshape_rectangular"]


@simple_block
def reshape_rectangular(x, tall=True):
    """ 
        If x is not 2D already, reshape it to 2D using the best
        factorization.    
        
        By default, the output is a "tall" array.
        
        Note that if the size of x is a prime number N,
        the output will be 1 x N.
    """
    if x.ndim == 2:
        return x

    shape = get_best_shape(x.size, tall=tall)
    x = x.reshape(shape)
    return x


def get_best_shape(size, tall=True):
    # Find all divisors
    divs = list(divisors(size))

    # Order this by similarity
    def score(x):
        """ 0 if square, positive if skewed. """
        a, b = x
        return np.abs(np.log(a) - np.log(b))

    divso = sorted(divs, key=score)

    # print('divisors: %s' % divso)
    shape = divso[0]

    it_is_tall = shape[0] > shape[1]
    transpose = tall ^ it_is_tall
    if transpose:
        shape = (shape[1], shape[0])

    return shape


def divisors(size):
    """ 
        Returns a list of (a, b) tuples such that a*b=c and a<=b.
        
        (a very naive implementation) 
    """
    L = []
    for i in range(1, size + 1):
        if size % i == 0:
            a, b = i, size / i
            L.append((min(a, b), max(a, b)))
    return list(set(L))



from .utils import PGTestCase

good_examples = [
    """
--- model master

|input name=a| -> |slave1| -> |slave2| -> |output name=b| 

--- model slave1

# |input name=x| -> |output name=y|
(x) -> (y)

--- model slave2

|input name=x| -> |output name=y|
""",
    """
# recursive models
--- model master

|input name=a| -> |output name=b| 

--- model slave

|input name=x| -> |output name=b|
""",
]

bad_examples = [
    # XXXXXXXXXXXXXXXXX how can it work?
    """
# recursive models
--- model master

|input name=a| -> |slave1| -> |output name=b| 

--- model slave1

|input name=x| -> |slave2| -> |output name=y|

--- model slave2

|input name=x| -> |slave1| -> |output name=y|
""",
    """
# same name should throw an error
--- model master

|input name=a| -> |slave| -> |output name=b| 

--- model slave

|input name=x|   -> |output name=y|

--- model slave

|input name=x|  -> |output name=y|
""",
]


class SemanticsTest(PGTestCase):
    def testExamples(self):
        for example in good_examples:
            model = self.check_semantic_ok(example)
            print(model)

    def testBadExamples(self):
        for example in bad_examples:
            self.check_semantic_error(example)



from procgraph import Block, ETERNITY
from procgraph.block_utils import make_sure_dir_exists
import pickle


class Pickle(Block):
    """ Dumps the input as a :py:mod:`pickle` file. """

    Block.alias("pickle")
    Block.config("file", "File to write to.")
    Block.input("x", "Anything pickable.")

    def write(self, x, filename):
        make_sure_dir_exists(filename)
        with open(filename, "wb") as f:
            pickle.dump(x, f, protocol=pickle.HIGHEST_PROTOCOL)

    def update(self):
        self.write(self.input.x, self.config.file + ".part")

    def finish(self):
        self.write(self.input.x, self.config.file)


class PickleLoad(Block):
    """ Dumps the input as a :py:mod:`pickle` file. """

    Block.alias("pickle_load")
    Block.config("file", "File to read from.")
    Block.output("x", "Object read from file")

    def update(self):
        with open(self.config.file, "rb") as f:
            x = pickle.load(f)
            self.set_output(0, x, timestamp=ETERNITY)


class PickleGroup(Block):
    """ Dumps the input as a :py:mod:`pickle` file, in the form
        of a dictionary  signal name -> value.    
    """

    Block.alias("pickle_group")
    Block.config("file", "File to write to.")
    Block.input_is_variable("Any number of pickable signals.")

    def write(self, x, filename):
        make_sure_dir_exists(filename)
        with open(filename, "wb") as f:
            pickle.dump(x, f, protocol=pickle.HIGHEST_PROTOCOL)

    def get_dict(self):
        data = {}
        for signal in self.get_input_signals_names():
            data[signal] = self.get_input(signal)
        return data

    def update(self):
        self.write(self.get_dict(), self.config.file + ".part")

    def finish(self):
        self.write(self.get_dict(), self.config.file)



from hdflog import PROCGRAPH_LOG_GROUP, check_is_procgraph_log, tc_open_for_reading, tc_close
from procgraph import Block, Generator, BadConfig
import glob
import numpy
import operator
import os


__all__ = ["HDFread_many"]


class HDFread_many(Generator):
    """
        This block is a variation on :ref:`block:hdfread` that can
        concatenate the output of logs stored in multiple HDF files.

        The files are specified using a wildcard: for example, ``dir/*.h5``.

        A difference with :ref:`block:hdfread` is that all signals
        must be specified explicitly (:ref:`block:hdfread` can guess); the
        reason is that we want to avoid reading unrelated logs with different
        signals.

        We check that all files specified have all required signals.

        The logfiles are read in the order compatible with their timestamp.
    """

    Block.alias("hdfread_many")
    Block.output_is_defined_at_runtime("The signals read from the logs.")

    Block.config("files", "HDF files to read; you can use the wildcard ``*``.")
    Block.config(
        "signals", "Which signals to output (and in what order). " "Should be a comma-separated list. "
    )

    Block.config("quiet", "If true, disables advancements status messages.", default=False)

    def get_output_signals(self):
        self.signals = list([x for x in self.config.signals.split(",") if x])
        if not self.signals:
            raise BadConfig("No signals specified.", self, "signals")
        return self.signals

    def init(self):
        # Find the required files
        files = glob.glob(self.config.files)

        if len(files) == 0:
            raise BadConfig("No files correspond to the pattern %r." % self.config.files, self, "files")

        # Open all logs; make sure they have the required signals and
        # note their initial timestamp.

        # list(tuple(file, timestamp, signal2table))
        logs = []
        for f in files:
            hf = tc_open_for_reading(f)
            check_is_procgraph_log(hf)

            log_group = hf.root._f_getChild(PROCGRAPH_LOG_GROUP)
            log_signals = list(log_group._v_children)

            signal2table = {}
            for s in self.signals:
                if not s in log_signals:
                    raise Exception("Log %r does not have signal %r " "(it has %s)." % (f, s, log_signals))
                signal2table[s] = log_group._f_getChild(s)

            timestamp = signal2table[s][0]["time"]
            logs.append((hf, timestamp, signal2table))

        # Sort them by timestamp
        self.logs = sorted(logs, key=operator.itemgetter(1))
        self.current_log = None

        for log in logs:
            filename = log[0].filename
            length = len(list(log[2].values())[0])
            self.status("- queued log (%6d rows) from %r" % (length, filename))

        self.start_reading_next_log()

    def status(self, s):
        if not self.config.quiet:
            self.info(s)

    def start_reading_next_log(self):
        if self.current_log is not None:
            self.status("Just finished log %r." % self.current_log[0].filename)
            tc_close(self.current_log[0])
            self.current_log = None

        if not self.logs:
            # we finished
            self.status("No logs remaining.")
            return

        self.current_log = self.logs.pop(0)
        self.signal2table = self.current_log[2]

        self.status("Now starting with %r." % self.current_log[0].filename)

        # signal -> index in the table (or None)
        self.signal2index = {}

        for signal in self.signals:
            if len(self.signal2table[signal]) > 0:
                self.signal2index[signal] = 0
            else:
                self.signal2index[signal] = None

    def _choose_next_signal(self):
        """ Returns a tuple (name,timestamp) of the signal that produces
            the next event, or (None,None) if we finished the log. """
        # array of tuples (signal, timestamp)
        status = []
        for signal in self.signals:
            index = self.signal2index[signal]
            if index is not None:
                table = self.signal2table[signal]
                timestamp = table[index]["time"]
                status.append((signal, timestamp))

        if not status:
            return (None, None)
        else:
            sorted_status = sorted(status, key=operator.itemgetter(1))
            return sorted_status[0]

    def next_data_status(self):
        next_signal, next_timestamp = self._choose_next_signal()
        if next_signal is None:
            # one log is finished:
            if not self.logs:
                # no more logs
                return (False, None)
            else:
                self.start_reading_next_log()
                return self.next_data_status()
        else:
            return (True, next_timestamp)

    def update(self):
        next_signal, next_timestamp = self._choose_next_signal()
        assert next_signal is not None

        # get value
        table = self.signal2table[next_signal]
        index = self.signal2index[next_signal]
        assert next_timestamp == table[index]["time"]
        value = table[index]["value"]

        self.set_output(next_signal, value=value, timestamp=next_timestamp)

        # update index
        if index + 1 == len(table):
            # finished
            self.status("Finished reading signal %r." % next_signal)
            self.signal2index[next_signal] = None
        else:
            self.signal2index[next_signal] = index + 1

        # Status messages for the first signal
        if next_signal == self.signals[0]:
            T = len(table)
            nintervals = 10
            interval = int(numpy.floor(T * 1.0 / nintervals))
            if index > 0 and index != interval * (nintervals) and index % interval == 0:
                percentage = index * 100.0 / T

                T = str(T)
                index = str(index).rjust(len(T))
                self.status(
                    "Read %.0f%% (%s/%s) of %r (tracking signal %r)."
                    % (percentage, index, T, os.path.basename(self.current_log[0].filename), next_signal)
                )

    def finish(self):
        self.start_reading_next_log()



from abc import abstractmethod
import string
import traceback
from types import GeneratorType

from contracts import describe_type, describe_value, contract
from contracts.utils import indent

from procgraph import Generator
from procgraph import ModelExecutionError


__all__ = [
    "IteratorGenerator",
]


class IteratorGenerator(Generator):
    """

    """

    @abstractmethod
    @contract(returns=GeneratorType)
    def init_iterator(self):
        """ Must return an iterator yielding signal, timestamp, value """
        pass

    def init(self):
        self.iterator = self.init_iterator()
        if self.iterator is None:
            msg = "must return an iterator, got %s" % describe_value(self.iterator)
            raise ValueError(msg)
        self._load_next()

    def _load_next(self):
        try:
            try:
                res = next(self.iterator)
                if not isinstance(res, tuple):
                    msg = "Expected tuple (signal, timestamp, value), obtained %s" % describe_type(res)
                    raise ValueError(msg)
                if not len(res) == 3:
                    raise ValueError("Required len 3 tuple; obtained %d." % len(res))
                signal, timestamp, value = res
            except StopIteration:
                raise
            except Exception as e:
                msg = "Could not call next() on user-given iterator.\n"
                msg += "   iterator: %s\n" % str(self.iterator)
                msg += "    of type: %s\n" % describe_type(self.iterator)
                msg += "because of this error:\n"
                msg += indent(string.strip("%s\n%s" % (e, traceback.format_exc())), "| ")
                raise ModelExecutionError(msg, self)

            if not isinstance(signal, (str, int)):
                msg = "Expected a string or number for the signal, got %s" % describe_value(signal)
                raise ValueError(msg)

            if not isinstance(timestamp, float):
                msg = "Expected a number for the timestamp, got %s" % describe_value(timestamp)
                raise ValueError(msg)

            self.next_signal = signal
            self.next_timestamp = timestamp
            self.next_value = value
            self.has_next = True
        except StopIteration:
            self.has_next = False

    def next_data_status(self):
        if self.has_next:
            return (True, self.next_timestamp)
        else:
            return (False, None)

    def update(self):
        if not self.has_next:
            return  # XXX: error here?

        self.set_output(self.next_signal, value=self.next_value, timestamp=self.next_timestamp)

        self._load_next()



""" 
    All the arbitrary constants used by ProcGraph.


    Note that some of these come from prototypes, and I should get rid of them.

    Also some are configuration parameters which should be settable in 
    some way.
    
"""
from contracts import new_contract


# If true, consider an error if a model's input is not defined formally.
STRICT_CHECK_OF_DEFINED_IO = False


# STRICT = True:
#
#  Unused variables trigger error
#
STRICT = False

# ## For the procgraph_info  dictionary:

# Variable that should be defined
PROCGRAPH_INFO_VARIABLE = "procgraph_info"
# Which variable holds the dependencies
REQUIRES = "requires"
# Key for the "parsed" version of REQUIRES (used internally)
REQUIRES_PARSED = "requires_parsed"


# ## Types of inputs and outputs

# Normal input/output
FIXED = "fixed-signal"
# Variable number of inputs/outputs.
VARIABLE = "variable-signal"
# The block will define the signals *after* it has seen the configuration.
DEFINED_AT_RUNTIME = "defined-at-runtime"

# ## When defining simple blocks
# TODO: write documentation on how these work
COMPULSORY = "compulsory-param"
REQUIRED = COMPULSORY

# if the result is this value, no output is generated
NO_OUTPUT = "no-output-generated"

TIMESTAMP = "timestamp-param"


# not sure this is used anywhere now
ETERNITY = "eternity"

# ## Environment

# Environment variable containing colon-separated additional paths to look
# for modules.
PATH_ENV_VAR = "PROCGRAPH_PATH"


@new_contract
def pg_timestamp_or_eternity(x):
    """ Either x is a float or x is equal to ETERNITY
        (ETERNITY = min R) """
    return x == ETERNITY or isinstance(x, float)



""" 
    This is a set of blocks to read and write logs in ROS_ Bag format.
    
    You need the `rospy` package to be installed.
    
    .. _pytables: http://pytables.org
     
    .. _ROS: http://www.ros.org

"""

procgraph_info = {
    # List of python packages
    "requires": ["ros", "rosbag_utils"]
}


# Smart dependency importing
from procgraph import import_magic, import_succesful, logger

rosbag = import_magic(__name__, "ros", "rosbag")
rosbag_utils = import_magic(__name__, "rosbag_utils")

if import_succesful(rosbag_utils):
    from .bagread import BagRead
    from .bagwrite import BagWrite
    from .conversions import *
else:
    msg = "Could not import rosbag_utils; install from git@github.com:AndreaCensi/ros_node_utils.git"
    logger.error(msg)

from procgraph import pg_add_this_package_models

pg_add_this_package_models(__file__, __package__)



from . import np

from procgraph import Block
from procgraph.block_utils import assert_rgb_image


class Compose(Block):
    """
    Compose several images in the same canvas.
    You should probably use :ref:`block:grid` in many situations.

    Example configuration: ::

        compose.positions = {y: [0,0], ys: [320,20]}

    """

    Block.alias("compose")

    Block.config("width", "Dimension in pixels.")
    Block.config("height", "Dimension in pixels.")
    Block.config("positions", "A structure giving the position of each signal in the canvas.")

    Block.input_is_variable("Images to compose.")

    Block.output("canvas", "RGB image")

    def update(self):
        width = self.get_config("width")
        height = self.get_config("height")

        canvas = np.zeros((height, width, 3), dtype="uint8")

        positions = self.get_config("positions")

        if not isinstance(positions, dict):
            raise Exception('I expected a dict, not "%s"' % positions)

        for signal, position in list(positions.items()):
            if not self.is_valid_input_name(signal):
                raise Exception('Unknown input "%s" in %s.' % (signal, self))
            rgb = self.get_input(signal)
            # TODO check
            if rgb is not None:
                assert_rgb_image(rgb, "input %s to compose block" % signal)

                place_at(canvas, rgb, position[0], position[1])
                # print "Writing image %s" % signal
            else:
                self.warning("Ignoring image %s because not ready.\n" % signal)

        self.set_output(0, canvas)


def place_at(canvas, image, xpix, ypix):
    # print canvas.shape, image.shape
    xsize = min(canvas.shape[1] - xpix, image.shape[1])
    ysize = min(canvas.shape[0] - ypix, image.shape[0])
    if len(image.shape) == 2:
        image = image.reshape((image.shape[0], image.shape[1], 1))
    canvas[ypix : (ypix + ysize), xpix : (xpix + xsize), 0:3] = image[0:ysize, 0:xsize, :]



import operator

from procgraph import Block, Generator, BadConfig

import numpy
from hdflog.tables_cache import tc_close

# TODO: check there is at least one signal

# TODO: respect original order


class HDFread(Generator):
    """
        This block reads a log written with :ref:`block:hdfwrite`.

    """

    Block.alias("hdfread")
    Block.output_is_defined_at_runtime("The signals read from the log.")
    Block.config("file", "HDF file to read")
    Block.config(
        "signals",
        "Which signals to output (and in what order). "
        "Should be a comma-separated list. If you do not specify it "
        " will be all signals (TODO: in the original order).",
        default=None,
    )

    Block.config("quiet", "If true, disables advancements status messages.", default=False)

    def get_output_signals(self):
        self.reader = self.config.file
        all_signals = self.reader.get_all_signals()

        if self.config.signals is None:
            self.signals = all_signals
        else:
            signal_list = list([x for x in self.config.signals.split(",") if x])
            if not signal_list:
                msg = "Bad format: %r." % self.config.signals
                raise BadConfig(msg, self, "signals")
            for s in signal_list:
                if not s in all_signals:
                    msg = "Signal %r not present in log (available: %r)" % (s, all_signals)
                    raise BadConfig(msg, self, "signals")
                self.signals.append(s)

        return self.signals

    def init(self):
        # let's do the rest of the initialization

        # signal -> table
        self.signal2table = {}
        # signal -> index in the table (or None)
        self.signal2index = {}

        for signal in self.signals:
            self.signal2table[signal] = self.log_group._f_getChild(signal)
            if len(self.signal2table[signal]) > 0:
                self.signal2index[signal] = 0
            else:
                self.signal2index[signal] = None

    def _choose_next_signal(self):
        """ Returns a tuple (name,timestamp) of the signal that produces
            the next event, or (None,None) if we finished the log. """
        # array of tuples (signal, timestamp)
        status = []
        for signal in self.signals:
            index = self.signal2index[signal]
            if index is not None:
                table = self.signal2table[signal]
                timestamp = table[index]["time"]
                status.append((signal, timestamp))

        if not status:
            return (None, None)
        else:
            sorted_status = sorted(status, key=operator.itemgetter(1))
            return sorted_status[0]

    def next_data_status(self):
        next_signal, next_timestamp = self._choose_next_signal()
        if next_signal is None:
            return (False, None)
        else:
            return (True, next_timestamp)

    def update(self):
        next_signal, next_timestamp = self._choose_next_signal()

        table = self.signal2table[next_signal]
        index = self.signal2index[next_signal]
        assert next_timestamp == table[index]["time"]
        value = table[index]["value"]

        self.set_output(next_signal, value=value, timestamp=next_timestamp)

        # update index
        if index + 1 == len(table):
            # finished
            self.signal2index[next_signal] = None
        else:
            self.signal2index[next_signal] = index + 1

        # write status message if not quiet
        if next_signal == self.signals[0] and not self.config.quiet:
            self.write_update_message(index, len(table), next_signal)

    def write_update_message(self, index, T, signal, nintervals=10):
        interval = int(numpy.floor(T * 1.0 / nintervals))
        if index > 0 and index != interval * (nintervals) and index % interval == 0:
            percentage = index * 100.0 / T
            T = str(T)
            index = str(index).rjust(len(T))
            self.debug(
                "%s read %.0f%% (%s/%s) (tracking signal %r)."
                % (self.config.file, percentage, index, T, signal)
            )

    def finish(self):
        tc_close(self.hf)



import traceback

from contracts import describe_type, indent, contract


__all__ = [
    "Library",
    "default_library",
]


class Library(object):
    def __init__(self, parent=None):
        self.parent = parent

        self.name2block = {}

        # Set of filenames already looked into
        self.loaded_files = set()

    def exists(self, block_type):
        if block_type in self.name2block:
            return True
        else:
            if self.parent is not None:
                return self.parent.exists(block_type)
            else:
                return False

    def register(self, block_type, generator):
        if self.exists(block_type):
            msg = "Type %r already registered." % block_type
            raise ValueError(msg)

        self.name2block[block_type] = generator

    def instance(self, block_type, name, config, library=None):
        from procgraph.core.exceptions import ModelInstantionError, SemanticError

        if library is None:
            library = self
        generator = self.get_generator_for_block_type(block_type)
        try:
            block = generator(name=name, config=config, library=library)
        except Exception as e:
            msg = "Could not instance block from generator.\n"
            msg += "        name: %s  \n" % name
            msg += "      config: %s  \n" % config
            msg += "   generator: %s  \n" % generator
            msg += "     of type: %s  \n" % describe_type(generator)
            msg += "Because of this exception:\n"
            if isinstance(e, (SemanticError, ModelInstantionError)):
                msg += indent("%s" % e, "| ")
            else:
                msg += indent("%s\n%s" % (e, traceback.format_exc()), "| ")
            raise ModelInstantionError(msg) from e  # TODO: use Procgraph own exception

        block.__dict__["generator"] = generator
        return block

    def get_generator_for_block_type(self, block_type):
        """ Returns the generator object for the block type.
        
            A generator can instance using:
            
                generator(name, config, library)
                
            and it can provide (before instancing) the properties:
            
                generator.config
                generator.input
                generator.output
        """

        if not self.exists(block_type):
            msg = "Block %r does not exist." % block_type

            msg += "\n Known: %s" % self.get_known_blocks()
            raise ValueError(msg)

        if block_type in self.name2block:
            generator = self.name2block[block_type]
            return generator
        else:
            assert self.parent
            return self.parent.get_generator_for_block_type(block_type)

    @contract(returns="list(str)")
    def get_known_blocks(self):
        blocks = list(self.name2block.keys())
        if self.parent:
            blocks.extend(self.parent.get_known_blocks())
        return blocks


default_library = Library()



from .utils import PGTestCase

good_examples = [
    """
--- model ciao
input x1
input x2 "output 2"
output o1 

"""
]

bad_examples = [
    """
# cannot have newline before doc string
--- model ciao
input x3 
"documentation"
""",
    """
# Names must be strings
--- model ciao
input 0 "not a string" 
""",
]


class SyntaxTestMultiple(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from .utils import PGTestCase

good_examples = [
    "x = 1",
    "x = [1]",
    "x = [1,2]",
    "x = {}",
    "x = {a: b}",
    'x = {"a": "b"}',
    'x = {"a": "b", a: d}',
    'x = {"a": "b", a: {a:1}}',
    "x = []",
    "x = [1]",
]


bad_examples = [
    # empty string gives error
    "x = [1",
    "x = [1 2]",
    "x = {0: a}",
]


class SyntaxTest2(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            print(("Trying: %s" % example))
            self.check_syntax_ok(example)



from .timestamps import TIMESTAMP_FIELD, iso_from_timestamp
from procgraph.utils import system_cmd_result, CmdException
import os
import yaml
from . import logger


def get_ffmpeg_metadata_args(md, timestamp):
    args = []
    if timestamp is not None:
        iso = iso_from_timestamp(timestamp)
        # timestamp_ff = str(dt)
        # args += ['-timestamp', timestamp_ff] # deprecated
        md["creation_time"] = iso
        md[TIMESTAMP_FIELD] = iso
    for k, v in list(md.items()):
        args += ["-metadata", "%s=%s" % (k, v)]
    return args


def get_metadata_file(filename):
    return filename + ".metadata.yaml"


def write_extra_metadata_for(filename, md):
    metadata_file = get_metadata_file(filename)
    with open(metadata_file, "w") as f:
        f.write(yaml.dump(md, default_flow_style=False))


def read_extra_metadata_for(filename):
    metadata_file = get_metadata_file(filename)
    if not os.path.exists(metadata_file):
        return dict(warning="Metadata file %s did not exist" % filename)
    with open(metadata_file, "r") as f:
        return yaml.load(f)


def ffmpeg_get_metadata(video):
    cmds = ["ffmpeg", "-i", video, "-f", "ffmetadata", "-"]
    try:
        res = system_cmd_result(
            ".",
            cmds,
            display_stdout=False,
            display_stderr=False,
            raise_on_error=True,
            capture_keyboard_interrupt=False,
        )
    except CmdException as e:
        # TODO: be more descriptive
        logger.debug(e)
        return {}

    lines = res.stdout.split("\n")
    assert lines[0] == ";FFMETADATA1"

    keys = {}
    for line in lines[1:]:
        line = line.strip()
        if not line:
            continue
        k, v = line.split("=")
        if v[0] == "'":
            v = v[1:-1]

        k = k.lower()
        keys[k] = v

    return dict(keys)



from hdflog import PGHDFLogWriter
from procgraph import Block, BadInput
import numpy


__all__ = ["HDFwrite"]


class HDFwrite(Block):
    """ This block writes the incoming signals to a file in HDF_ format.
     
    The HDF format is organized as follows: ::
    
         /            (root)
         /procgraph             (group with name procgraph)
         /procgraph/signal1     (table)
         /procgraph/signal2     (table)
         ...
         
    Each table has the following fields:
    
         time         (float64 timestamp)
         value        (the datatype of the signal)
         
    If a signal changes datatype, then an exception is raised.
    
    """

    Block.alias("hdfwrite")
    Block.input_is_variable("Signals to be written", min=1)
    Block.config("file", "HDF file to write")
    Block.config("compress", "Whether to compress the hdf table.", 1)
    Block.config("complib", "Compression library (zlib, bzip2, blosc, lzo).", default="zlib")
    Block.config("complevel", "Compression level (0-9)", 9)

    def init(self):
        self.writer = PGHDFLogWriter(
            self.config.file,
            compress=self.config.compress,
            complevel=self.config.complevel,
            complib=self.config.complib,
        )

    def update(self):
        signals = self.get_input_signals_names()
        for signal in signals:
            if self.input_update_available(signal):
                self.log_signal(signal)

    def log_signal(self, signal):
        timestamp = self.get_input_timestamp(signal)
        value = self.get_input(signal)
        # only do something if we have something
        if value is None:
            return
        assert timestamp is not None

        if not isinstance(value, numpy.ndarray):
            # TODO: try converting
            try:
                value = numpy.array(value)
            except:
                msg = "I can only log numpy arrays, not %r" % value.__class__
                raise BadInput(msg, self, signal)

        self.writer.log_signal(timestamp, signal, value)

    def finish(self):
        self.writer.finish()



from collections import namedtuple
from procgraph import Block, ETERNITY, Generator


__all__ = [
    "Sync",
]


Sample = namedtuple("Sample", "timestamp value")


class Sync(Generator):
    """
    This block synchronizes a set of streams to the first stream (the master).

    The first signal is called the "master" signal.
    The other (N-1) are slaves.

    We guarantee that:

    - if the slaves are faster than the master,
      then we output exactly the same.

    Example diagrams: ::

        Master  *  *  *   *   *
        Slave   ++++++++++++++++

        Master  *  *  *   *   *
        output? v  v  x   v
        Slave   +    +      +
        output? v    v      v
    """

    Block.alias("sync")

    Block.input_is_variable("Signals to synchronize. The first is the master.", min=2)
    Block.output_is_variable("Synchronized signals.")

    def init(self):
        # output signals get the same name as the inputs
        names = self.get_input_signals_names()

        # create a state for each signal: it is an array
        # of tuples (timestamp, tuple)
        queues = {}
        for signal in names:
            queues[signal] = []

        # note in all queues,
        # [(t0,...), (t1,...), ... , (tn,...) ]
        # and we have t0 > tn
        # The chronologically first (oldest) is queue[-1]
        # You get one out using queue.pop()
        # You insert one with queue.insert(0,...)

        self.set_state("queues", queues)
        self.state.already_seen = {}

        self.set_state("master", names[0])
        self.set_state("slaves", names[1:])

        # The output is an array of tuple (timestamp, values)
        # [
        #    (timestamp1, [value1,value2,value3,...]),
        #    (timestamp1, [value1,value2,value3,...])
        # ]#
        self.set_state("output", [])

    def update(self):
        def debug(s):
            if False:  # XXX: use facilities
                print(("sync %s %s" % (self.name, s)))

        output = self.get_state("output")
        queues = self.get_state("queues")
        names = self.get_input_signals_names()
        # for each input signal, put its value in the queues
        # if it is not already present
        for i, name in enumerate(names):
            #            Commenting this; we clarified 0 = eternity; None = no signal yet
            if not self.input_signal_ready(i):
                debug("Ignoring signal %r because timestamp is None." % name)
                continue

            current_timestamp = self.get_input_timestamp(i)
            current_value = self.get_input(i)

            if name in self.state.already_seen and self.state.already_seen[name] == current_timestamp:
                continue
            else:
                self.state.already_seen[name] = current_timestamp

            queue = queues[name]
            # if there is nothing in the queue
            # or this is a new sample
            if (len(queue) == 0) or newest(queue).timestamp != current_timestamp:  # new sample
                debug("Inserting signal '%s'  ts %s (queue len: %d)" % (name, current_timestamp, len(queue)))
                # debug('Before the queue is: %s' % queue)
                add_last(queue, Sample(timestamp=current_timestamp, value=current_value))

                # debug('Now the queue is: %s' % queue)

        master = self.get_state("master")
        master_queue = queues[master]
        slaves = self.get_state("slaves")

        # if there is more than one value in each slave
        if len(master_queue) > 1:
            val = master_queue.pop()
            debug("DROPPING master (%s) ts =%s" % (master, val.timestamp))

        # Now check whether all slaves signals are >= the master
        # If so, output a synchronized sample.
        if master_queue:
            all_ready = True
            master_timestamp = master_queue[-1].timestamp
            # print "Master timestamp: %s" % (master_timestamp)
            for slave in slaves:
                slave_queue = queues[slave]
                # remove oldest
                while (
                    len(slave_queue) > 1
                    and oldest(slave_queue).timestamp < master_timestamp
                    and oldest(slave_queue).timestamp != ETERNITY
                ):
                    debug("DROP one from %s" % slave)
                    slave_queue.pop()

                if not slave_queue:
                    # or (master_timestamp > slave_queue[-1].timestamp):
                    all_ready = False
                    # its_ts = map(lambda x:x.timestamp, slave_queue)
                    # print "Slave %s not ready: %s" %(slave, its_ts)
                    break

            if all_ready:
                debug("**** Ready: master timestamp %s " % (master_timestamp))
                master_value = master_queue.pop().value
                output_values = [master_value]
                for slave in slaves:
                    # get the freshest still after the master
                    slave_queue = queues[slave]
                    slave_timestamp, slave_value = slave_queue.pop()
                    # print('Slave, %s %s ' % (slave, slave_timestamp))
                    if slave_timestamp == ETERNITY:
                        slave_queue.append((slave_timestamp, slave_value))
                    # # not true anymore, assert slave_timestamp >=
                    # master_timestamp
                    # difference = slave_timestamp - master_timestamp
                    # debug(" - %s timestamp %s diff %s" %
                    #      (slave, slave_timestamp, difference))
                    output_values.append(slave_value)
                output.insert(0, (master_timestamp, output_values))

        # XXX XXX not really sure here
        # if master has more than one sample, then drop the first one

        # if we have something to output, do it
        if output:
            timestamp, values = output.pop()
            # FIXME: had to remove this for Vehicles simulation
            # assert timestamp > 0
            debug("---------------- @ %s" % timestamp)
            for i in range(self.num_output_signals()):
                self.set_output(i, values[i], timestamp)

    def next_data_status(self):
        output = self.get_state("output")
        if not output:  # no output ready
            return (False, None)
        else:
            timestamp = self.output[-1][0]
            return (True, timestamp)


def oldest(queue):
    return queue[-1]


def newest(queue):
    return queue[0]


def add_last(queue, ob):
    queue.insert(0, ob)



""" 

    Some functions specific to robotics applications. 

    Requires: http://github.com/AndreaCensi/geometry
    
"""


procgraph_info = {
    # List of python packages
    "requires": ["geometry"]
}

# Smart dependency resolution
from procgraph import import_magic, import_succesful

geometry = import_magic(__name__, "geometry")

if import_succesful(geometry):
    from . import pose2velocity
    from . import laser_display
    from . import laser_dot_display
    from . import organic_scale
    from . import misc
    from . import pose2velocity_b



import os

from setuptools import find_packages, setup


def get_version(filename):
    import ast

    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename="src/procgraph/__init__.py")

# Load autogenerated list of packages and dependencies
# from procgraph_packages import index

index = {
    "packages": {
        "procgraph.components.debug_components": {
            "blocks": ["info", "constant", "+", "*", "-", "/", "gain", "print", "identity"],
            "desc": "Components used for debugging and unit tests.",
            "requires": {},
        },
        "procgraph_cv": {
            "blocks": ["gradient", "smooth"],
            "desc": "Blocks using the OpenCV library.",
            "requires": {"cv": ["cv", "opencv"]},
        },
        "procgraph_foo": {
            "blocks": ["block_example"],
            "desc": "An example package for ProcGraph that shows how to organize your code. ",
            "requires": {"pickle": ["cPickle", "pickle"]},
        },
        "procgraph_hdf": {
            "blocks": ["hdfread_many", "hdfread_test", "hdfread_many_test", "hdfwrite", "hdfread"],
            "desc": "This is a set of blocks to read and write logs in HDF_ format. ",
            "requires": {"tables": ["tables"]},
        },
        "procgraph_images": {
            "blocks": [
                "rgb2gray",
                "compose",
                "reshape2d",
                "solid",
                "blend",
                "skim_top_and_bottom",
                "posterize",
                "gray2rgb",
                "scale",
                "grid",
                "skim_top",
                "posneg",
                "border",
            ],
            "desc": "Blocks for basic operations on images. ",
            "requires": {},
        },
        "procgraph_io_misc": {
            "blocks": ["pickle_load", "to_file", "pickle", "pickle_group", "as_json"],
            "desc": "Miscellaneous functions to be better organized.",
            "requires": {"json": ["simplejson"]},
        },
        "procgraph_mpl": {
            "blocks": ["plot"],
            "desc": "Blocks using Matplotlib to display data.",
            "requires": {"matplotlib": ["matplotlib"], "matplotlib.pylab": ["matplotlib.pylab"]},
        },
        "procgraph_mplayer": {
            "blocks": ["mplayer", "mencoder"],
            "desc": "Blocks for encoding/decoding video based on MPlayer.",
            "requires": {},
        },
        "procgraph_numpy_ops": {
            "blocks": [
                "smooth1d",
                "square",
                "sign",
                "minimum",
                "select",
                "rad2deg",
                "outer",
                "log",
                "sum",
                "astype",
                "arctan",
                "abs",
                "take",
                "real",
                "deg2rad",
                "hstack",
                "flipud",
                "max",
                "vstack",
                "gradient1d",
                "dstack",
                "maximum",
                "fliplr",
                "normalize_Linf",
                "mean",
            ],
            "desc": "Various operations wrapping numpy functions.",
            "requires": {},
        },
        "procgraph_pil": {
            "blocks": ["imread", "text", "resize"],
            "desc": "Blocks for image operations based on the PIL library",
            "requires": {"PIL": ["PIL"]},
        },
        "procgraph_robotics": {
            "blocks": [
                "pose2commands",
                "laser_display",
                "skim",
                "organic_scale",
                "pose2vel_",
                "laser_dot_display",
            ],
            "desc": "Some functions specific to robotics applications. ",
            "requires": {"geometry": ["geometry"]},
        },
        "procgraph_ros": {
            "blocks": ["bagread_test", "ros2python", "bagread"],
            "desc": "This is a set of blocks to read and write logs in ROS_ Bag format. ",
            "requires": {"ros": ["ros"]},
        },
        "procgraph_signals": {
            "blocks": [
                "derivative2",
                "slice",
                "two_step_difference",
                "derivative",
                "join",
                "sync",
                "fps_data_limit",
                "history",
                "historyt",
                "forward_difference",
                "low_pass",
                "make_tuple",
                "fps_limit",
                "fps_print",
                "last_n_samples",
                "sieve",
                "extract",
                "any",
                "wait",
            ],
            "desc": "Blocks performing operations with a dynamic nature. ",
            "requires": {},
        },
        "procgraph_statistics": {
            "blocks": ["cov2corr", "normalize", "expectation", "covariance", "soft_variance", "variance"],
            "desc": "Blocks for common statistical operations.",
            "requires": {},
        },
        "procgraph_yaml": {
            "blocks": ["yaml2object"],
            "desc": "YAML conversions.",
            "requires": {"yaml": ["yaml"]},
        },
    }
}


def is_package_available(p):
    """ Checks that a package is available. """
    try:
        __import__(p)
        return True
    except ImportError as e:
        # print("Warning: cannot import %r: %s" % (p, e))
        return False


missing = {}
problems = set()
for module, info in list(index["packages"].items()):
    for requirement, options in list(info["requires"].items()):
        if not any([is_package_available(op) for op in options]):
            print(("For package %r, dependency %r cannot be satisfied." % (module, requirement)))
            missing[requirement] = options
            # XXX: this is slightly incorrect if different packages want
            # the same module but give different alternatives.
            problems.add(module)

if missing:
    print("\n\n")
    print("Dependency search problems summary")
    print("----------------------------------\n")
    print("I could not find the following packages installed:\n")
    for requirement, options in list(missing.items()):
        if len(options) > 1:
            possible = " or ".join([x.__repr__() for x in options])
            o = "(Satisfiable by %s.)" % possible
        else:
            o = ""
        print((" - %-20s  %s" % (requirement, o)))

    print("\nThese missing requirements could make the following packages " "not work properly:\n")

    for p in problems:
        desc = index["packages"][p]["desc"]
        print((" - %-20s  (%s)" % (p, desc)))

    print(
        "\nI will go ahead and install everything, but you should install the missing \n"
        "packages for maximum functionality. An error will be thrown when you actually \n"
        "try to use the blocks in those packages. \n"
    )
    # if os.isatty(0):
    # 	raw_input('             Press any key to continue...')
    print("\n\n")

# check we are not forgetting new packages that weren't included in the
# autogenerated list yet.
packages = find_packages(where="src")
for p in packages:
    if p == "procgraph" or p.startswith("procgraph."):
        continue
    if not p in index["packages"]:
        print(("Warning: packages %r is not in the autogenerated index." % p))


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


line = "z6"

install_requires = [
    "setuptools-git",
    "pyparsing",
    "setproctitle",
    "termcolor",
    f"SystemCmd-{line}",
    f"PyGeometry-{line}",
    f"zuper-commons-{line}",
    "PyContracts3",
    "qtfaststart",
    "numpy",
]

setup(
    name=f"procgraph-{line}",
    author="Andrea Censi",
    author_email="censi@mit.edu",
    url="http://andreacensi.github.com/procgraph/",
    description='ProcGraph ("processing graph") is a Python library '
    "for rapid prototyping of processing pipelines for "
    "logged and realtime data.",
    long_description=read("README.txt"),
    keywords="",
    license="LGPL",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Topic :: Scientific/Engineering",
    ],
    version=version,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    #       package_data={'': ['*.pg'],
    # 				'prograph_cv': ['models/cv_demo_feature.pg']},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    # TODO: add many depedendencies
    tests_require=["nose>1.1.2,<2"],
    entry_points={
        "console_scripts": [
            "pg = procgraph.scripts.pgmain:main",
            "pgdoc = procgraph.scripts.pgdoc:main",
            "pgindex = procgraph.scripts.pgindex:main",
            # some other utils
            "pg-video-info = procgraph_mplayer.scripts.video_info:main",
            "pg-video-convert = procgraph_mplayer.scripts.video_convert:main",
            "pg-video-crop = procgraph_mplayer.scripts.crop_video:main",
            "pg-video-join = procgraph_mplayer.scripts.join_video:main",
            "pg-video-background = procgraph_mplayer.scripts.find_background:main",
            "pg-video-tomp4 = procgraph_mplayer.scripts.tomp4:main",
        ]
    },
    download_url=("http://github.com/AndreaCensi/procgraph/tarball/%s" % version),
)



from . import pylab
from six import StringIO
import numpy as np


__all__ = ["pylab2rgb"]


def pylab2rgb(transparent=False, tight=False):
    """ Saves and returns the pixels in the current pylab figure. 
    
        Returns a RGB uint8 array. Uses PIL to do the job.
        
        If transparent is true, returns a RGBA image instead of RGB. 
    """
    imgdata = StringIO()

    if tight:
        pylab.savefig(imgdata, format="png", bbox_inches="tight", pad_inches=0)
    else:
        pylab.savefig(imgdata, format="png")

    from procgraph_pil import Image

    imgdata.seek(0)
    im = Image.open(imgdata)
    if not transparent:
        im = im.convert("RGB")
    rgb = np.asarray(im)

    if transparent:
        rgba = rgb.copy()
        white = np.logical_and(rgb[:, :, 0] == 255, rgb[:, :, 1] == 255, rgb[:, :, 2] == 255)
        alpha = rgba[:, :, 3]
        alpha[white] = 0
        # alpha[numpy.logical_not(white)] = 110
        rgba[:, :, 3] = alpha
        return rgba

    return rgb



""" Miscellaneous functions to be better organized. """

procgraph_info = {"requires": [("json", ("simplejson",))]}  # TODO: add cjson?


from procgraph import import_magic

json = import_magic(__name__, "json")


from . import json_misc
from . import to_file
from . import pickling



import os
import fnmatch
import inspect
import pickle

from .model_instantiation import create_from_parsing_results
from .visualization import warning, debug
from .parsing import parse_model, ParsedModel
from .exceptions import SemanticError
from .registrar import default_library, Library
from .constants import PATH_ENV_VAR
from .. import deny_pgc_cache


class ModelSpec(object):
    """ Class used to register as a block type """

    def __init__(self, parsed_model, defined_in):
        self.parsed_model = parsed_model

        # each generator should expose this
        self.config = self.parsed_model.config
        self.input = self.parsed_model.input
        self.output = self.parsed_model.output

        # the module to which this model is associated
        self.defined_in = defined_in
        assert defined_in is not None

    def __call__(self, name, config, library):
        parsed_model = self.parsed_model

        parent = self

        # We create a mock library that forbids that this
        # model is created again. This prevents recursion.
        class ForbidRecursion(Library):
            def __init__(self, parent, forbid):
                Library.__init__(self, parent)
                self.forbid = forbid

            def instance(self, block_type, name, config, parent_library=None):
                if block_type == self.forbid:
                    msg = "Recursion error for model %r." % self.forbid
                    raise SemanticError(msg, parent.parsed_model)
                else:
                    return Library.instance(self, block_type, name, config, parent_library)

            def get_generator_for_block_type(self, block_type):
                if block_type == self.forbid:
                    msg = "Recursion error for model %r." % self.forbid
                    raise SemanticError(msg, parent.parsed_model)
                else:
                    return Library.get_generator_for_block_type(self, block_type)

        sandbox = ForbidRecursion(library, parsed_model.name)
        model = create_from_parsing_results(parsed_model, name, config, library=sandbox)

        return model


def pg_add_this_package_models(filename, assign_to, subdir="models"):
    """ Add the models for this package.
        Shortcut to put into the module ``__init__.py``.
        Call with filename = __file__, assign_to= __package__.

        Example: ::

            pg_add_this_package_models(filename__file__, assign_to=__package__)

    """

    if subdir is not None:
        dirname = os.path.join(os.path.dirname(filename), subdir)
    else:
        dirname = os.path.dirname(filename)

    pg_look_for_models(
        default_library, additional_paths=[dirname], ignore_env=True, assign_to_module=assign_to
    )


def pg_look_for_models(
    library, additional_paths=None, ignore_env=False, ignore_cache=False, assign_to_module=None
):
    """ Call this function at the beginning of the executions.
    It scans the disk for model definitions, and register
    them as available block types.
    Other than the paths that are passed by argument,
    it looks into the ones in the PROCGRAPH_PATH environment
    variable (colon separated list of paths), unless ignore_env is True.

    assign_to_module is a string that gives the nominal module the model
    is associated to -- this is only used for the documentation generation.

    TODO: add global cache in user directory.

    Honors global deny_pgc_cache to disable all caches.

    """

    paths = []
    if additional_paths:
        paths.extend(additional_paths)

    if not ignore_env:
        if PATH_ENV_VAR in os.environ:
            paths.extend(os.environ[PATH_ENV_VAR].split(":"))

    if not paths:
        if False:
            # TODO: add verbose switch
            warning("No paths given and environment var %r not defined." % PATH_ENV_VAR)

    # enumerate each sub directory
    all_files = set()
    for path in paths:
        if not os.path.isdir(path):
            # XXX: should I use exception?
            raise Exception("Invalid path %r to search for models. " % path)

        for root, dirs, files in os.walk(path):  # @UnusedVariable
            for f in files:
                if fnmatch.fnmatch(f, "*.pg"):
                    all_files.add((path, os.path.join(root, f)))

    for path, f in all_files:
        split = os.path.splitext(os.path.basename(f))
        base = split[0]

        # logger.debug('Loading models from %r' % f)

        # Make sure we use an absolute filename
        f = os.path.realpath(f)

        if f in library.loaded_files:
            # Skip if already looked at it
            continue

        library.loaded_files.add(f)

        cache = os.path.splitext(f)[0] + ".pgc"

        if deny_pgc_cache:
            # global switch
            ignore_cache = True

        if not ignore_cache and os.path.exists(cache) and os.path.getmtime(cache) > os.path.getmtime(f):
            try:
                models = pickle.load(open(cache))
            except Exception as e:
                debug("Cannot unpickle file %r: %s" % (cache, e))
                # XXX repeated code
                # debug("Parsing %r." % os.path.relpath(f))
                model_spec = open(f).read()

                models = parse_model(model_spec, filename=f)

        else:
            # debug("Parsing %r." % os.path.relpath(f))
            model_spec = open(f).read()
            models = parse_model(model_spec, filename=f)

        try:
            # TODO: make it parallel
            with open(cache, "w") as f:
                pickle.dump(models, f)
        except Exception as e:
            # Cannot write on the cache for whatever reason
            # debug('Cannot write cache file: %s' % e)
            try:
                if os.path.exists(cache):
                    os.unlink(cache)
            except:
                pass

        if models[0].name is None:
            models[0].name = base

        for parsed_model in models:
            if library.exists(parsed_model.name):
                if parsed_model.name in library.name2block:
                    prev = library.name2block[parsed_model.name].parsed_model.where
                else:
                    prev = "?"
                msg = "Found model %r in file:\n %r, and already in\n %r. " % (
                    parsed_model.name,
                    f,
                    prev.filename,
                )
                raise SemanticError(msg, parsed_model)

            if assign_to_module is None:
                assign_to_module = path
            model_spec = ModelSpec(parsed_model, assign_to_module)

            library.register(parsed_model.name, model_spec)


def pg_add_parsed_model_to_library(parsed_model, library, defined_in):
    assert parsed_model.name is not None
    if library.exists(parsed_model.name):
        prev = library.name2block[parsed_model.name].parsed_model.where
        # FIXME: suppose that I have tutorials.pg
        #  and I do:
        #    pg -d . tutorials.pg
        # This will fail because it will try to read tutorials.pg twice
        msg = "I already have registered model %r from %r. " % (parsed_model.name, prev.filename)
        raise SemanticError(msg, parsed_model)

    model_spec = ModelSpec(parsed_model, defined_in)
    library.register(parsed_model.name, model_spec)


def add_models_to_library(library, string, name=None, filename=None, defined_in=None):
    """
    defined_in: module NAME (to display in documentation)
        (Compulsory!)
    """
    if filename is None and defined_in is not None:
        filename = __import__(defined_in, fromlist=["x"]).__file__

    models = parse_model(string, filename=filename)
    if models[0].name is None:
        assert name is not None
        models[0].name = name

    for model in models:
        pg_add_parsed_model_to_library(parsed_model=model, library=library, defined_in=defined_in)


def model_from_string(model_spec, name=None, config=None, library=None, filename=None):
    """ Instances a model from a specification. Optional
        attributes can be passed. Returns a Model object.

        Additional models in the spec (after the first) are automatically
        added to the library (defined_in = calling module).

    """
    if config is None:
        config = {}
    if library is None:
        library = default_library
    assert isinstance(model_spec, str)
    assert isinstance(config, dict)
    assert name is None or isinstance(name, str)

    parsed_models = parse_model(model_spec, filename)

    assert isinstance(parsed_models, list)
    for x in parsed_models:
        assert isinstance(x, ParsedModel)

    # Get the first one
    parsed_model = parsed_models[0]

    if len(parsed_models) > 1:
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])

        for support in parsed_models[1:]:
            pg_add_parsed_model_to_library(support, library, defined_in=mod)

            # TODO: introduce the concept of submodels?
            # support.parent = parsed_model
            # parsed_model.children.append(support)

    model = create_from_parsing_results(parsed_model, name=name, config=config, library=library)

    return model



""" 
    Blocks for common statistical operations.

"""
from procgraph import import_magic


procgraph_info = {
    # List of python packages
    "requires": ["astatsa", "numpy"]
}

np = import_magic(__name__, "numpy")

from . import expectation
from . import variance
from . import covariance
from . import cov2corr
from . import minimum
from . import covariance2



import numpy

from procgraph import simple_block


@simple_block
def gradient1d(a):
    """
        Computes the gradient of a 1D array.

        :param a: Numpy array
        :type a: array(N),N>3

        :return: typed: The gradient of the array.
        :rtype: array
    """
    a = numpy.array(a)

    if len(a.shape) != 1 or len(a) < 3:
        raise ValueError("Expected 1D array, got shape %s" % str(a.shape))

    b = numpy.ndarray(shape=a.shape, dtype=a.dtype)

    n = len(a)
    for i in range(1, n - 1):
        b[i] = int((a[i + 1] - a[i - 1]) / 2)

    b[0] = b[1]
    b[-1] = b[-2]

    return b



# ---------------------------------------------------------
# natsort.py: Natural string sorting.
# ---------------------------------------------------------

# By Seo Sanghyeon.  Some changes by Connelly Barnes.
# Some more changes (key) by Andrea Censi.


def try_int(s):
    "Convert to integer if possible."
    try:
        return int(s)
    except:
        return s


def natsort_key(s):
    "Used internally to get a tuple by which s is sorted."
    import re

    return list(map(try_int, re.findall(r"(\d+|\D+)", s)))


def natcmp(a, b):
    "Natural string comparison, case sensitive."
    return cmp(natsort_key(a), natsort_key(b))


def natcasecmp(a, b):
    "Natural string comparison, ignores case."
    return natcmp(a.lower(), b.lower())


def natsort(seq, kcmp=natcmp, key=None):
    "In-place natural string sort."
    seq.sort(kcmp, key)


def natsorted(seq, kcmp=natcmp, key=None):
    "Returns a copy of seq, sorted by natural string sort."
    # convert set -> list
    temp = list(seq)
    natsort(temp, kcmp, key)
    return temp



from ..core.docstring_parsing import DocStringInfo, Arg, parse_docstring_annotations

examples = {
    """
        Provides a RGB representation of the values by interpolating the range
        [min(value),max(value)] into the colorspace [min_color, max_color].

        :param value: The field to represent.
        :type  value: HxW array

        :param max_value: If specified, everything *above* is clipped.
        :type max_value: float
        :param min_value: If specified, everything *below* is clipped.
        :type min_value: float

        :param min_color:  Color to give to the minimum values.


        :return: A RGB image.
        :rtype: HxWx3 uint8

        :return: gray

    """: DocStringInfo(
        docstring=(
            " \n        Provides a RGB representation"
            " of the values by interpolating the range \n"
            "        [min(value),max(value)] into "
            "the colorspace [min_color, max_color].\n"
        ),
        params={
            "value": Arg("The field to represent.", "HxW array"),
            "max_value": Arg("If specified, everything *above* is clipped.", "float"),
            "min_value": Arg("If specified, everything *below* is clipped.", "float"),
            "min_color": Arg("Color to give to the minimum values.", None),
        },
        returns=[Arg("A RGB image.", "HxWx3 uint8"), Arg("gray", None)],
    )
}


from .utils import PGTestCase


class DocStringTest(PGTestCase):
    def test_parsing(self):
        for string, result in list(examples.items()):
            parsed = parse_docstring_annotations(string)
            # print parsed
            # print result
            self.assertEqual(result, parsed)



from procgraph import Block


class ASync(Block):
    """ 
      The first signal is the "master".
      Waits that all signals are perceived once.
      Then it creates one event every time the master arrives.
    """

    Block.alias("async")

    Block.input_is_variable("Signals to (a)synchronize. The first is the master.", min=2)
    Block.output_is_variable("Synchronized signals.")

    def init(self):
        self.state.last_sent_timestamp = None

    def update(self):
        self.state.last_timestamp = None

        # No signal until everybody is ready
        if not self.all_input_signals_ready():
            return

        # No signal unless the master is ready
        master_timestamp = self.get_input_timestamp(0)
        if self.state.last_timestamp == master_timestamp:
            return

        self.state.last_timestamp = master_timestamp

        # Write all signals using master's timestamp
        for s in self.get_input_signals_names():
            self.set_output(s, self.get_input(s), master_timestamp)



import time
import traceback
import warnings

from typing import List, Optional

from .block import Block, Generator
from .constants import ETERNITY, STRICT_CHECK_OF_DEFINED_IO
from .exceptions import BadMethodCall, ModelExecutionError, ModelWriterError, SemanticError
from .model_io import ModelInput, ModelOutput
from .model_stats import ExecutionStats, write_stats
from .visualization import debug as debug_main, info, warning
from ..utils import indent


try:
    from time import thread_time as measure_thread_time
except ImportError:
    from time import clock as measure_thread_time

__all__ = ["Model"]


class Model(Generator):
    """
        A Model is a block and a generator.


        cleanup() and finish(): "cleanup" is the emergency clean up.
        It is called after finish().

        Order in which init(), update(), cleanup() can be called:
        - init(), update(), ..., update(), finish(), cleanup()
        - init(), finish(), cleanup()
        - init(), cleanup()

    """

    def __init__(self, name, model_name):
        """ Name is the personal name of this instance.
            model_name is this model's name.

            model_name is used for visualization and serialization
            config is usef only for serialization.
             """
        if name is None:
            name = "unnamed-block"
        assert isinstance(name, str)

        if model_name is None:
            model_name = "unnamed-model"
        assert isinstance(model_name, str)

        self.model_name = model_name

        # As a block
        Block.__init__(self, name=name, config={}, library=None)

        self.name2block = {}
        self.name2block_connection = {}
        # block -> block unresolved
        self.unresolved = {}

        # list of blocks that act as generators (instance of Generator)
        self.generators = []

        self.reset_execution()

        # dict signal name -> Block for blocks of type ModelInput
        self.model_input_ports = {}
        # dict signal name -> Block for blocks of type ModelInput
        self.model_output_ports = {}

        self.stats = ExecutionStats()

        self._already_inited = False

        self.level = 0

    def add_block(self, name, block):
        """  Add a block to the model. Returns the block instance. """
        assert not name in self.name2block, "Name %r already taken" % name

        self.name2block[name] = block

        if isinstance(block, Generator):
            self.generators.append(block)

        if isinstance(block, ModelInput):
            block.init()  # get the name from config
            if not self.is_valid_input_name(block.signal_name):
                msg = "Input %r was not defined formally." % block.signal_name
                if STRICT_CHECK_OF_DEFINED_IO:
                    raise SemanticError(msg, block)
                else:
                    warning("Warning: %s" % msg)

                if self.are_input_signals_defined():
                    all_inputs = self.get_input_signals_names()
                else:
                    all_inputs = []

                self.define_input_signals_new(all_inputs + [block.signal_name])

            self.model_input_ports[block.signal_name] = block

        if isinstance(block, ModelOutput):
            block.init()  # get the name from config

            if not self.is_valid_output_name(block.signal_name):
                msg = "Output %r was not defined formally." % block.signal_name
                if STRICT_CHECK_OF_DEFINED_IO:
                    raise SemanticError(msg, block)
                else:
                    warning("Warning: %s" % msg)

                if self.are_output_signals_defined():
                    all_outputs = self.get_output_signals_names()
                else:
                    all_outputs = []

                # XXX: bug: output_signals -> output_signals
                self.define_output_signals_new(all_outputs + [block.signal_name])

            self.model_output_ports[block.signal_name] = block

        return block

    def from_outside_set_input(self, num_or_id, value, timestamp):
        assert self._already_inited, "The block must first be init()ed."
        Block.from_outside_set_input(self, num_or_id, value, timestamp)

        signal_name = self.canonicalize_input(num_or_id)
        input_block = self.model_input_ports[signal_name]
        input_block.set_output(signal_name, value, timestamp)
        self.blocks_to_update.append(input_block)

    def connect(self, block1, block1_signal, block2, block2_signal, public_name):
        """ Caller should check that the public name is not taken. """
        assert public_name is not None
        if public_name in self.public_signal_names():
            msg = "The name %r is already present in:\n" % public_name
            for s in self.public_signal_names():
                msg += "- %s\n" % s
            raise ValueError(msg)

        BC = BlockConnection(block1, block1_signal, block2, block2_signal, public_name)

        self.name2block_connection[public_name] = BC

    def public_signal_names(self):
        """ Returns the set of public signal names. """
        return set(self.name2block_connection.keys())

    def next_data_status(self):
        """ XXX: OK, I'm writing this late and probably it's more
            complicated than this. """
        generator_timestamps = []
        at_least_one = False
        for generator in self.generators:
            status = generator.next_data_status()
            (has_next, timestamp) = status  # @UnusedVariable
            if has_next:
                at_least_one = True
                if timestamp is not None:
                    generator_timestamps.append(timestamp)
                else:
                    pass
        #                    self.debug('No timestamp for %s' % generator)

        if not at_least_one:
            return False, None
        elif not generator_timestamps:
            return True, None
        else:
            return True, min(generator_timestamps)

    def has_more(self):
        """ Returns true if there are blocks with pending updates,
            or there is at least one generator that has not ended. """
        if self.blocks_to_update:
            return True

        for generator in self.generators:
            # print('checking %s' % generator)
            status = generator.next_data_status()  # @UnusedVariable

            if not isinstance(status, tuple) or len(status) != 2:
                msg = "next_data_status() should return a tuple of len 2, " " not %r." % status
                raise ModelWriterError(msg, generator)
            (has_next, timestamp) = status  # @UnusedVariable
            if has_next:
                return True

        return False

    def reset_execution(self):
        self.blocks_to_update = []
        # add all the blocks without input to the update list
        for block in list(self.name2block.values()):
            if isinstance(block, Model):
                # Don't consider generators because has_more() will consider
                # them.
                # This is mainly for propagating constants.
                if block.blocks_to_update:
                    self.blocks_to_update.append(block)
                pass
            elif (
                not isinstance(block, ModelInput)
                and not isinstance(block, Generator)
                and block.num_input_signals() == 0
            ):
                # These are the constants blocks.
                self.blocks_to_update.append(block)

    def init(self):
        assert not self._already_inited, "The block has already been init()ed."
        self._already_inited = True
        for block in list(self.name2block.values()):
            block.level = self.level + 1
            block.init()
        self.reset_execution()

    def finish(self):
        for block in list(self.name2block.values()):
            try:
                block.finish()
            except BaseException:
                raise BadMethodCall("finish", block, traceback.format_exc())
            except BadMethodCall as e:
                e.blocks.insert(0, block)
                raise e

    def cleanup(self):
        """ We try hard to call cleanup() on all the blocks. """
        blocks_failed = []
        msg = ""
        for block in self.name2block.values():
            try:
                block.cleanup()
            except Exception as e:
                msg += "Cleanup for %s failed:\n%s\n" % (block, indent(e, "> "))
                blocks_failed.append(block)

        if blocks_failed:
            s = "Cleanup for %d blocks failed.\n%s" % (len(blocks_failed), msg)
            raise Exception(s)  # XXX: which other exception?

    def update(self):
        do_stats = True

        if do_stats:
            pass
            # warnings.warn("Performance much reduced due to stats.")

        # Turn on debugging here
        def debug(s):
            if False:
                debug_main(f"Model {self.model_name} | {s}")

        # We keep a list of blocks to be updated.
        # If the list is not empty, then pop one and update it.
        if self.blocks_to_update:
            # get one block
            block = self.blocks_to_update.pop(0)

            debug("Got block to update %s" % block)

        else:
            debug("No blocks to update")

            # look if we have any generators
            # list of (generator, timestamp)
            generators_with_timestamps = []
            for generator in self.generators:
                (has_next, timestamp) = generator.next_data_status()
                if has_next:
                    generators_with_timestamps.append((generator, timestamp))

            if not generators_with_timestamps:
                msg = "You asked me to update but nothing's left."
                raise ModelExecutionError(msg, self)

            # now look for the smallest available timestamp
            # (timestamp can be none)
            def cmpk(timestamp1, timestamp2):
                if timestamp1 is None:
                    return 1
                elif timestamp2 is None:
                    return -1
                elif timestamp1 < timestamp2:
                    return -1
                elif timestamp2 < timestamp1:
                    return 1
                else:
                    return 0

            # XXX: assuming that key had the priority
            generators_with_timestamps.sort(key=lambda x: x[1])  # , key=functools.cmp_to_key(cmpk))
            block = generators_with_timestamps[0][0]

        if block is None:
            # We finished everything
            msg = "You asked me to update but nothing's left."
            raise ModelExecutionError(msg, self)

        # now we have a block (could be a generator)
        #        debug('Updating %s (input ts: %s)' %
        #              (block, block.get_input_signals_timestamps()))

        # We also time the execution
        start_cpu = measure_thread_time()
        start_wall = time.time()

        try:
            result = block.update()
        except BadMethodCall as e:
            e.blocks.insert(0, block)
            raise e
        except Exception:
            raise BadMethodCall("update", block, traceback.format_exc())

        input_timestamps: List[Optional[float]] = block.get_input_signals_timestamps()
        input_timestamps_not_none: List[float] = list([_f for _f in input_timestamps if _f is not None])
        output_timestamps: List[Optional[float]] = block.get_output_signals_timestamps()

        try:
            if input_timestamps_not_none:
                timestamp = max(input_timestamps_not_none)
            elif output_timestamps:
                timestamp = max(output_timestamps)
            else:  # for those that don't have input signals
                timestamp = ETERNITY
        except BaseException as e:
            msg = f"i: {input_timestamps} o: {output_timestamps}"
            raise ValueError(msg) from e

        if do_stats:
            wall = time.time() - start_wall
            stop_cpu = measure_thread_time()
            cpu = stop_cpu - start_cpu
            if cpu == 0:
                msg = (
                    f"Low resolution clock() - start: {start_cpu} stop: {stop_cpu} wall: {wall} - using "
                    f"wall time."
                )
                print(msg)
                cpu = wall
                # raise ValueError('No CPU? wall = %s' % wall)

            # print('block: %s cpu: %s time: %s of %s' % (block, cpu, timestamp, block.get_output_signals_timestamps()))
            self.stats.add(block=block, cpu=cpu, wall=wall, timestamp=timestamp)

        # if the update is not finished, we put it back in the queue
        if result == block.UPDATE_NOT_FINISHED:
            self.blocks_to_update.insert(0, block)
        else:
            # the block updated, propagate
            # debug("  processed %s, ts: %s" % (block, block.get_output_signals_timestamps()))
            # debug("  its successors: %s" % list(self.__get_output_connections(block)))

            # check if the output signals were updated
            for connection in self.__get_output_connections(block):
                other = connection.block2
                # debug("  - considering connection %s to %s" % (connection, other))
                # Don't include dummy connection
                if other is None:
                    # debug("   ignoring dummy connection %s" % connection)
                    continue
                other_signal = connection.block2_signal

                this_signal = connection.block1_signal
                this_timestamp = block.get_output_timestamp(this_signal)

                value = block.get_output(this_signal)

                if value is not None and this_timestamp is None:
                    msg = (
                        "Strange, value is not None but the timestamp is 0"
                        " for output signal %r of block %s." % (block.canonicalize_output(this_signal), block)
                    )
                    raise ModelExecutionError(msg, block)

                # Ignore if this signal wasn't updated yet
                if this_timestamp is None:
                    # debug("   ignoring because timestamp is None")
                    continue

                to_update = (not other.input_signal_ready(other_signal)) or (
                    this_timestamp > other.get_input_timestamp(other_signal)
                )

                if to_update:
                    # debug('  then waking up %s' % other)

                    other.from_outside_set_input(other_signal, value, this_timestamp)

                    if not other in self.blocks_to_update:
                        self.blocks_to_update.append(other)

                    if do_stats:
                        self.stats.add_connection_sample(
                            connection=connection, value=value, timestamp=this_timestamp
                        )

                    # If this is an output port, update the model
                    if isinstance(other, ModelOutput):
                        self.set_output(other.signal_name, value, this_timestamp)
                else:
                    # debug("   not updated %s because not %s > %s." %
                    #       (other, this_timestamp, other.get_input_timestamp(other_signal)))
                    pass

        # now let's see if we have still work to do
        # this step is important when the model is inside another one
        if self.blocks_to_update:
            return Block.UPDATE_NOT_FINISHED
        else:
            return True

    def __get_output_connections(self, block):
        for block_connection in list(self.name2block_connection.values()):
            if block_connection.block1 == block:
                yield block_connection

    def __get_successors(self, block):
        """ Returns an iterable of all the blocks connected
            to one of the outputs of the given block. """
        successors = set()
        for block_connection in list(self.name2block_connection.values()):
            if block_connection.block1 == block:
                successors.add(block_connection.block2)
        return successors

    def __repr__(self):
        s = "M:%s:%s(" % (self.model_name, self.name)
        s += self.get_io_repr()
        s += ")"
        return s

    def summary(self):
        info("--- Model: %d blocks, %d connections" % (len(self.name2block), len(self.name2block_connection)))
        for name, block in list(self.name2block.items()):
            info("- %s: %s" % (name, block))

        for name, conn in list(self.name2block_connection.items()):
            info("- %s: %s" % (name, conn))

    def print_stats(self):
        """ Prints statistics for the leaves of the hierarchy. """
        all_samples = self.collect_stats_leaves()
        write_stats(all_samples)

    def collect_stats_leaves(self):
        """ Collects the stats from the leaves. """
        all_samples = []
        for block in list(self.name2block.values()):
            if isinstance(block, Model):
                all_samples.extend(block.collect_stats_leaves())

        leaves = [v for (b, v) in list(self.stats.samples.items()) if not isinstance(b, Model)]
        all_samples.extend(leaves)
        return all_samples

    def stats_get_all(self):
        """ Returns a structure for all stats recursively. """
        s = self.stats.get_all()
        for block in list(self.name2block.values()):
            if isinstance(block, Model):
                block_stats = block.stats_get_all()
                block_blocks = block_stats["blocks"]
                for bname, x in list(block_blocks.items()):
                    s["blocks"]["%s.%s" % (block.name, bname)] = x

                block_signals = block_stats["signals"]
                for signal in block_signals:
                    s2 = signal.copy()
                    s2["block1"] = "%s.%s" % (block.name, s2["block1"])

                    s2["block2"] = "%s.%s" % (block.name, s2["block2"])
                    s["signals"].append(s2)

        return s


class BlockConnection(object):
    def __init__(self, block1, block1_signal, block2, block2_signal, public_name=None):
        assert isinstance(block1, Block)
        assert block1_signal is not None
        assert block2 is None or isinstance(block2, Block)

        self.block1 = block1
        self.block1_signal = block1_signal
        self.block2 = block2
        self.block2_signal = block2_signal
        self.public_name = public_name

    def __repr__(self):
        s = "Connection("
        s += self.block1.name
        s += ".%s" % self.block1_signal

        s += " --> "
        if self.block2:
            s += self.block2.name
            s += ".%s" % self.block2_signal
        else:
            s += "?.?"
        s += ")"

        return s



from .conversions import pg_video_convert
from .scripts.crop_video import video_crop
from contextlib import contextmanager
from procgraph import Block
from procgraph.block_utils import expand, make_sure_dir_exists, input_check_rgb_or_grayscale
from procgraph.utils import friendly_path, indent
import numpy
import os
import signal
import subprocess
import tempfile


__all__ = ["MEncoder"]


class MEncoder(Block):
    """
        Encodes a video stream using ``mencoder``.

        Note that allowed codec and bitrate depend on
        your version of mencoder.

        MP4 output: currently it works by creating a .AVI with mencoder
        and then converting to MP4 using ffmpeg.

        RGBA videos: not working fully. For now it outputs .AVI with
        mencoder and encoded using png.


    """

    Block.alias("mencoder")

    Block.input(
        "image",
        "Either a HxWx3 uint8 numpy array representing " "an RGB image, or a HxW representing grayscale. ",
    )

    Block.config("file", "Output file (AVI/MP4 format)")
    Block.config(
        "fps",
        "Framerate of resulting movie. If not specified, " "it will be guessed from data.",
        default=None,
    )
    Block.config(
        "fps_safe",
        "If the frame autodetect gives strange results, " "we use this safe value instead.",
        default=10,
    )

    Block.config("quiet", "If True, suppress mencoder's messages", default=True)
    Block.config(
        "timestamps",
        "If True, also writes <file>.timestamps that" " includes a line with the timestamp for" " each frame",
        default=True,
    )

    Block.config("crop", "If true, the video will be " "post-processed and cropped", default=False)

    Block.config("md", "Metadata for the file.", default={})

    Block.config("container", "Which container to use; " "if None, it will be guessed.", default=None)
    Block.config("vcodec", "Codec to use. If None, it will be guessed", default=None)
    Block.config("vcodec_params", "Codec-dependent params.", default={})

    Block.config("firstpass_bitrate", default=3 * 1000 * 1000)

    def init(self):
        self.process = None
        self.buffer = []
        self.image_shape = None  # Shape of image being encoded

        self.first_frame_timestamp = None

        from .programs_existence import check_programs_existence

        programs = ["mencoder", "ffmpeg"]
        self.programs = check_programs_existence(programs=programs)

        for p in self.programs:
            self.info("Using %13s = %s" % (p, self.programs[p]))

    def update(self):
        input_check_rgb_or_grayscale(self, 0)

        # Put image in a buffer -- we don't use it right away
        image = self.get_input(0)

        timestamp = self.get_input_timestamp(0)
        self.buffer.append((timestamp, image))

        if self.process is None:
            self.try_initialization()

        if self.process is not None:
            # initialization was succesful
            while self.buffer:
                timestamp, image = self.buffer.pop(0)

                if self.first_frame_timestamp is None:
                    self.first_frame_timestamp = timestamp

                self.write_value(timestamp, image)

    def try_initialization(self):
        # If we don't have at least two frames, continue
        if len(self.buffer) < 2:
            return

        # Get height and width from first image
        first_image = self.buffer[0][1]

        self.shape = first_image.shape
        self.height = self.shape[0]
        self.width = self.shape[1]

        if self.height > 8192 or self.width > 8192:
            msg = "Mencoder cannot support movies this big (%sx%s)"
            msg = msg % (self.height, self.width)
            raise Exception(msg)
        self.ndim = len(self.shape)

        # Format for mencoder's rawvideo "format" option
        if self.ndim == 2:
            format = "y8"  # @ReservedAssignment
        else:
            if self.shape[2] == 3:
                format = "rgb24"  # @ReservedAssignment
            elif self.shape[2] == 4:
                # Note: did not try this yet
                format = "rgba"  # @ReservedAssignment
                msg = "I detected that you are trying to write a transparent"
                msg += "video. This does not work well yet (and besides,"
                msg += "it is not supported in many applications, like "
                msg += "Keynote). Anyway, the plan is to use mencoder "
                msg += 'to write a .AVI with codec "png". This will fail '
                msg += "for now, but perhaps in the future it will be "
                msg += ".better"
                self.error(msg)

        # guess the fps if we are not given the config
        if self.config.fps is None:
            delta = self.buffer[-1][0] - self.buffer[0][0]

            if delta == 0:
                timestamps = [x[0] for x in self.buffer]
                self.debug("Got 0 delta: timestamps: %s" % timestamps)
                fps = 0
            else:
                fps = (len(self.buffer) - 1) / delta

            # Check for very wrong results
            if not (3 < fps < 60):
                self.error(
                    "Detected fps is %.2f; this seems strange to me,"
                    " so I will use the safe choice fps = %.2f." % (fps, self.config.fps_safe)
                )
                fps = self.config.fps_safe
        else:
            fps = self.config.fps

        # adapt the bitrate to the size of the image
        vbitrate0 = self.config.firstpass_bitrate
        shape0 = (640, 480)

        n1 = self.width * self.height
        n0 = shape0[0] * shape0[1]
        vbitrate = vbitrate0 * n1 / n0

        self.info("Estimated bitrate %r" % vbitrate)

        max_bitrate = 90 * 1000 * 1000
        if vbitrate > max_bitrate:

            vbitrate = max_bitrate
            self.info("Estimated bitrate too high, capping at %s" % vbitrate)

        self.filename = expand(self.config.file)
        if os.path.exists(self.filename):
            self.info("Removing previous version of %s." % friendly_path(self.filename))
            os.unlink(self.filename)

        self.tmp_filename = f"{self.filename}-active.avi"

        make_sure_dir_exists(self.filename)

        self.info(
            "Writing %dx%d %s video stream at %.3f fps to %r."
            % (self.width, self.height, format, fps, friendly_path(self.filename))
        )

        if format == "rgba":
            ovc = ["-ovc", "lavc", "-lavcopts", "vcodec=png"]
        else:
            ovc = ["-ovc", "lavc", "-lavcopts", "vcodec=%s:vbitrate=%d" % ("mpeg4", vbitrate)]

        out = ["-o", self.tmp_filename]
        args = (
            [
                self.programs["mencoder"],
                "/dev/stdin",
                "-demuxer",
                "rawvideo",
                "-rawvideo",
                "w=%d:h=%d:fps=%f:format=%s" % (self.width, self.height, fps, format),
            ]
            + ovc
            + out
        )

        # '-v', "0", # verbosity level (1 prints stats \r)

        # self.debug('$ %s' % " ".join(args))
        # Note: mp4 encoding is currently broken in mencoder :-(
        #       so we have to use ffmpeg as a second step.
        # These would be the options to add:
        # '-of', 'lavf', '-lavfopts', 'format=mp4'

        self.tmp_stdout = tempfile.TemporaryFile()
        self.tmp_stderr = tempfile.TemporaryFile()

        quiet = self.config.quiet
        if quiet:
            # XXX /dev/null not portable
            #            self.debug('stderr: %s' % self.tmp_stderr)
            #            self.debug('stdout: %s' % self.tmp_stdout)
            # TODO: write error
            self.process = subprocess.Popen(
                args,
                preexec_fn=ignore_sigint,
                stdin=subprocess.PIPE,
                stdout=self.tmp_stdout.fileno(),
                stderr=self.tmp_stderr.fileno(),
            )
        else:
            self.process = subprocess.Popen(args=args, stdin=subprocess.PIPE)

        if self.config.timestamps:
            self.timestamps_filename = self.filename + ".timestamps"
            self.timestamps_file = open(self.timestamps_filename, "w")

    def _get_metadata(self):
        """ Returns the user-given metadata as well as some extra created by us. """
        metadata = self.config.md
        # metadata['author'] = 'procgraph'
        # metadata['show'] = 'procgraph'
        # metadata['title'] = 'A video by me.'
        # metadata['comment'] = 'Another video by me.'
        return metadata

    def finish(self):
        if self.process is None:
            msg = "Finish() before starting to encode."
            self.error(msg)
            raise Exception(msg)
            return

        timestamp = self.first_frame_timestamp
        metadata = self._get_metadata()
        container = self.config.container
        vcodec = self.config.vcodec
        vcodec_params = self.config.vcodec_params

        self.info("Transcoding %s" % friendly_path(self.filename))
        pg_video_convert(
            self.tmp_filename,
            self.filename,
            container=container,
            vcodec=vcodec,
            vcodec_params=vcodec_params,
            timestamp=timestamp,
            metadata=metadata,
        )

        #         if False:  # XXX
        if os.path.exists(self.tmp_filename):
            os.unlink(self.tmp_filename)

        if True:
            T = self.first_frame_timestamp
            os.utime(self.filename, (T, T))

        self.info("Finished %s" % friendly_path(self.filename))

        # TODO: skip mp4
        if self.config.crop:
            base, ext = os.path.splitext(self.filename)
            cropped = "%s-crop%s" % (base, ext)
            video_crop(self.filename, cropped)
            os.rename(cropped, self.filename)

    def cleanup(self):
        # TODO: remove timestamps
        #         if 'timestamps_filename' in self.__dict__:
        #             if os.path.exists(self.timestamps_filename):
        #                 os.unlink(self.timestamps_filename)

        if "tmp_filename" in self.__dict__:
            #             if False:  # XXX
            if os.path.exists(self.tmp_filename):
                os.unlink(self.tmp_filename)

        self.cleanup_mencoder()

    def cleanup_mencoder(self):
        # Try to cleanup as well as possible
        if (not "process" in self.__dict__) or self.process is None:
            return

        self.process.stdin.close()

        try:
            if False:
                with timeout(5):
                    self.process.terminate()
                    self.process.wait()
            else:
                self.process.terminate()
                self.process.wait()
        except Timeout as e:
            self.error(e)
        except (OSError, AttributeError):
            # Exception AttributeError: AttributeError("'NoneType' object
            # has no attribute 'SIGTERM'",)
            # in <bound method RangefinderUniform.__del__
            # of RangefinderUniform> ignored
            # http://stackoverflow.com/questions/2572172/
            pass

    #            self.error('Error while cleanup: %s' % e)

    def write_value(self, timestamp, image):
        if self.image_shape is None:
            self.image_shape = image.shape

        if self.image_shape != image.shape:
            msg = "The image has changed shape, from %s to %s." % (self.image_shape, image.shape)
            raise Exception(msg)  # TODO: badinput

        # very important! make sure we are using a reasonable array
        if not image.flags["C_CONTIGUOUS"]:
            image = numpy.ascontiguousarray(image)  # @UndefinedVariable

        try:
            if False:
                with timeout(5):
                    self.process.stdin.write(image.data)
                    self.process.stdin.flush()
            else:
                self.process.stdin.write(image.data)
                self.process.stdin.flush()
        except (Exception, KeyboardInterrupt) as e:
            # IOError = broken pipe
            msg = "Could not write data to mencoder: %s." % e
            #            self.error(msg)
            #            msg += '\n' + indent(self.process.stdout.read(), 'stdout> ')
            #            msg += '\n' + indent(self.process.stderr.read(), 'stderr> ')

            def read_all(f):
                os.lseek(f.fileno(), 0, 0)
                return f.read()

            stderr = read_all(self.tmp_stderr)
            stdout = read_all(self.tmp_stdout)

            msg += indent("stdout>", stderr)
            msg += indent("stderr>", stdout)

            raise Exception(msg)

        if self.config.timestamps:
            self.timestamps_file.write("%.4f\n" % timestamp)
            self.timestamps_file.flush()


class Timeout(Exception):
    pass


@contextmanager
def timeout(delta=5):
    def handler(signum, frame):
        raise Timeout("operation took more than %s seconds" % delta)

    # Set the signal handler and a 5-second alarm
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(delta)
    try:
        yield
    finally:
        signal.alarm(0)


def ignore_sigint():
    signal.signal(signal.SIGINT, signal.SIG_IGN)



from .utils import PGTestCase

good_examples = [
    """--- model ciao
y = 2
""",
    """------ model ciao
y = 2
""",
    """
--- model ciao
y = 2
--- model belle
y = 2
""",
    """

--- model ciao

y = 2

--- model belle
y = 2


""",
    """
# comments at the beginning should not start a model
--- model belle
e = 2
""",
    """
# spacing 
---   model   belle
e = 2
""",
    # Testing docstrings
    """
--- model ciao
''' This is the documentation
    for this model.            '''
y = 2
""",
    """

--- model ciao

''' This is the documentation
    for this model.            '''
y = 2
""",
    """
--- model ciao
config ciao "test"

''' this is a documentation for the block '''
|block|

""",
]

bad_examples = [
    """
# Mixing

y = 2

--- model belle
y = 2
""",
    """
# Empty
--- model belle

--- model belle
y = 2
""",
    """
# incomplete
--- model belle
e = 2
--- model belle

""",
    """
# bad name
--- model 1belle
e = 2
""",
    """
# bad syntax
-- model 1belle
e = 2
""",
    """
# should be on the same line 
---   model  
belle
e = 2
""",
    # docstring should be the first thing
    """
--- model ciao
y = 2
''' This is the documentation
    for this model.            '''

""",
]


class SyntaxTestMultiple(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from .utils import PGTestCase

good_examples = [
    (
        """
--- model master
config  a      "well documented param"
config  b = 3  "well documented param"

|verify x=$a y=$b|
""",
        {"a": 3, "b": 3},
    ),
    (
        """
--- model master
config  a      "well documented param"
config  b = 3  "well documented param"

|verify x=$a y=$b|
""",
        {"a": 3},
    ),
    # Now using the dict syntax with "."
    (
        """
--- model master
|verify y=2|
""",
        {"verify.x": 2},
    ),
    # Same thing, but with module name
    (
        """
--- model master
|v:verify y=2|
""",
        {"v.x": 2},
    ),
]

good_examples_plus = [
    # Now using the dict syntax without "."
    (
        """
--- model master
|v:verify y=2|
""",
        {"v": {"x": 2}},
    ),
    # Now using the dict syntax without "."
    (
        """
--- model master
|verify y=2|
""",
        {"verify": {"x": 2}},
    ),
]

bad_examples = [
    (
        """
# a is not assigned
--- model master
config  a      "well documented param"
config  b = 3  "well documented param"

|verify x=$a y=$b|
""",
        {"b": 3},
    ),
    (
        """
# double description
--- model master
config  a      "well documented param"
config  a      "well documented param"

b=2
""",
        {},
    ),
    (
        """
# overwriting configuration
--- model master
config  a      "well documented param"

a=2
""",
        {},
    ),
    (
        '''
--- model master
""" Cannot have a config variable with the same name as a block """
config  a      "well documented param"
|a:verify x=$a y=$a|
''',
        {},
    ),
]


class SemanticsTest(PGTestCase):

    #    @unittest.skip('This is not implemented yet')
    #    @unittest.expectedFailure
    #    def testNewExamples(self):
    #        for example, config in good_examples_plus:
    #            self.check_semantic_ok(example, config=config)

    def testExamples(self):
        for example, config in good_examples:
            self.check_semantic_ok(example, config=config)

    def testBadExamples(self):
        for example, config in bad_examples:
            self.check_semantic_error(example, config=config)
